import os
import shutil
import tempfile

import pytest

import salt.utils.files
import salt.utils.stringutils
from tests.support.mock import mock_open, patch
from tests.support.unit import TestCase

try:
    import M2Crypto

    HAS_M2 = True
except ImportError:
    HAS_M2 = False
try:
    from Cryptodome.PublicKey import RSA  # pylint: disable=unused-import

    HAS_PYCRYPTO_RSA = True
except ImportError:
    HAS_PYCRYPTO_RSA = False
if not HAS_PYCRYPTO_RSA:
    try:
        from Crypto.PublicKey import RSA  # nosec

        HAS_PYCRYPTO_RSA = True
    except ImportError:
        HAS_PYCRYPTO_RSA = False


PRIVKEY_DATA = (
    "-----BEGIN RSA PRIVATE KEY-----\n"
    "MIIEpAIBAAKCAQEA75GR6ZTv5JOv90Vq8tKhKC7YQnhDIo2hM0HVziTEk5R4UQBW\n"
    "a0CKytFMbTONY2msEDwX9iA0x7F5Lgj0X8eD4ZMsYqLzqjWMekLC8bjhxc+EuPo9\n"
    "Dygu3mJ2VgRC7XhlFpmdo5NN8J2E7B/CNB3R4hOcMMZNZdi0xLtFoTfwU61UPfFX\n"
    "14mV2laqLbvDEfQLJhUTDeFFV8EN5Z4H1ttLP3sMXJvc3EvM0JiDVj4l1TWFUHHz\n"
    "eFgCA1Im0lv8i7PFrgW7nyMfK9uDSsUmIp7k6ai4tVzwkTmV5PsriP1ju88Lo3MB\n"
    "4/sUmDv/JmlZ9YyzTO3Po8Uz3Aeq9HJWyBWHAQIDAQABAoIBAGOzBzBYZUWRGOgl\n"
    "IY8QjTT12dY/ymC05GM6gMobjxuD7FZ5d32HDLu/QrknfS3kKlFPUQGDAbQhbbb0\n"
    "zw6VL5NO9mfOPO2W/3FaG1sRgBQcerWonoSSSn8OJwVBHMFLG3a+U1Zh1UvPoiPK\n"
    "S734swIM+zFpNYivGPvOm/muF/waFf8tF/47t1cwt/JGXYQnkG/P7z0vp47Irpsb\n"
    "Yjw7vPe4BnbY6SppSxscW3KoV7GtJLFKIxAXbxsuJMF/rYe3O3w2VKJ1Sug1VDJl\n"
    "/GytwAkSUer84WwP2b07Wn4c5pCnmLslMgXCLkENgi1NnJMhYVOnckxGDZk54hqP\n"
    "9RbLnkkCgYEA/yKuWEvgdzYRYkqpzB0l9ka7Y00CV4Dha9Of6GjQi9i4VCJ/UFVr\n"
    "UlhTo5y0ZzpcDAPcoZf5CFZsD90a/BpQ3YTtdln2MMCL/Kr3QFmetkmDrt+3wYnX\n"
    "sKESfsa2nZdOATRpl1antpwyD4RzsAeOPwBiACj4fkq5iZJBSI0bxrMCgYEA8GFi\n"
    "qAjgKh81/Uai6KWTOW2kX02LEMVRrnZLQ9VPPLGid4KZDDk1/dEfxjjkcyOxX1Ux\n"
    "Klu4W8ZEdZyzPcJrfk7PdopfGOfrhWzkREK9C40H7ou/1jUecq/STPfSOmxh3Y+D\n"
    "ifMNO6z4sQAHx8VaHaxVsJ7SGR/spr0pkZL+NXsCgYEA84rIgBKWB1W+TGRXJzdf\n"
    "yHIGaCjXpm2pQMN3LmP3RrcuZWm0vBt94dHcrR5l+u/zc6iwEDTAjJvqdU4rdyEr\n"
    "tfkwr7v6TNlQB3WvpWanIPyVzfVSNFX/ZWSsAgZvxYjr9ixw6vzWBXOeOb/Gqu7b\n"
    "cvpLkjmJ0wxDhbXtyXKhZA8CgYBZyvcQb+hUs732M4mtQBSD0kohc5TsGdlOQ1AQ\n"
    "McFcmbpnzDghkclyW8jzwdLMk9uxEeDAwuxWE/UEvhlSi6qdzxC+Zifp5NBc0fVe\n"
    "7lMx2mfJGxj5CnSqQLVdHQHB4zSXkAGB6XHbBd0MOUeuvzDPfs2voVQ4IG3FR0oc\n"
    "3/znuwKBgQChZGH3McQcxmLA28aUwOVbWssfXKdDCsiJO+PEXXlL0maO3SbnFn+Q\n"
    "Tyf8oHI5cdP7AbwDSx9bUfRPjg9dKKmATBFr2bn216pjGxK0OjYOCntFTVr0psRB\n"
    "CrKg52Qrq71/2l4V2NLQZU40Dr1bN9V+Ftd9L0pvpCAEAWpIbLXGDw==\n"
    "-----END RSA PRIVATE KEY-----"
)

PUBKEY_DATA = (
    "-----BEGIN PUBLIC KEY-----\n"
    "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA75GR6ZTv5JOv90Vq8tKh\n"
    "KC7YQnhDIo2hM0HVziTEk5R4UQBWa0CKytFMbTONY2msEDwX9iA0x7F5Lgj0X8eD\n"
    "4ZMsYqLzqjWMekLC8bjhxc+EuPo9Dygu3mJ2VgRC7XhlFpmdo5NN8J2E7B/CNB3R\n"
    "4hOcMMZNZdi0xLtFoTfwU61UPfFX14mV2laqLbvDEfQLJhUTDeFFV8EN5Z4H1ttL\n"
    "P3sMXJvc3EvM0JiDVj4l1TWFUHHzeFgCA1Im0lv8i7PFrgW7nyMfK9uDSsUmIp7k\n"
    "6ai4tVzwkTmV5PsriP1ju88Lo3MB4/sUmDv/JmlZ9YyzTO3Po8Uz3Aeq9HJWyBWH\n"
    "AQIDAQAB\n"
    "-----END PUBLIC KEY-----"
)

MSG = b"It's me, Mario"

SIG = (
    b"\x07\xf3\xb1\xe7\xdb\x06\xf4_\xe2\xdc\xcb!F\xfb\xbex{W\x1d\xe4E"
    b"\xd3\r\xc5\x90\xca(\x05\x1d\x99\x8b\x1aug\x9f\x95>\x94\x7f\xe3+"
    b"\x12\xfa\x9c\xd4\xb8\x02]\x0e\xa5\xa3LL\xc3\xa2\x8f+\x83Z\x1b\x17"
    b'\xbfT\xd3\xc7\xfd\x0b\xf4\xd7J\xfe^\x86q"I\xa3x\xbc\xd3$\xe9M<\xe1'
    b"\x07\xad\xf2_\x9f\xfa\xf7g(~\xd8\xf5\xe7\xda-\xa3Ko\xfc.\x99\xcf"
    b"\x9b\xb9\xc1U\x97\x82'\xcb\xc6\x08\xaa\xa0\xe4\xd0\xc1+\xfc\x86"
    b'\r\xe4y\xb1#\xd3\x1dS\x96D28\xc4\xd5\r\xd4\x98\x1a44"\xd7\xc2\xb4'
    b"]\xa7\x0f\xa7Db\x85G\x8c\xd6\x94!\x8af1O\xf6g\xd7\x03\xfd\xb3\xbc"
    b"\xce\x9f\xe7\x015\xb8\x1d]AHK\xa0\x14m\xda=O\xa7\xde\xf2\xff\x9b"
    b"\x8e\x83\xc8j\x11\x1a\x98\x85\xde\xc5\x91\x07\x84!\x12^4\xcb\xa8"
    b"\x98\x8a\x8a&#\xb9(#?\x80\x15\x9eW\xb5\x12\xd1\x95S\xf2<G\xeb\xf1"
    b"\x14H\xb2\xc4>\xc3A\xed\x86x~\xcfU\xd5Q\xfe~\x10\xd2\x9b"
)


class TestBadCryptodomePubKey(TestCase):
    """
    Test that we can load public keys exported by pycrpytodome<=3.4.6
    """

    TEST_KEY = (
        "-----BEGIN RSA PUBLIC KEY-----\n"
        "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAzLtFhsvfbFDFaUgulSEX\n"
        "Gl12XriL1DT78Ef2/u8HHaSMmPie37BLWas/zaHwI6066bIyYQJ/nUCahTaoHM7L\n"
        "GlWc0wOU6zyfpihCRQHil05Y6F+olFBoZuYbFPvtp7/hJx/D7I/0n2o/c7M5i3Y2\n"
        "3sBxAYNooIQHXHUmPQW6C9iu95ylZDW8JQzYy/EI4vCC8yQMdTK8jK1FQV0Sbwny\n"
        "qcMxSyAWDoFbnhh2P2TnO8HOWuUOaXR8ZHOJzVcDl+a6ew+medW090x3K5O1f80D\n"
        "+WjgnG6b2HG7VQpOCfM2GALD/FrxicPilvZ38X1aLhJuwjmVE4LAAv8DVNJXohaO\n"
        "WQIDAQAB\n"
        "-----END RSA PUBLIC KEY-----\n"
    )

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.key_path = os.path.join(self.test_dir, "cryptodom-3.4.6.pub")
        with salt.utils.files.fopen(self.key_path, "wb") as fd:
            fd.write(self.TEST_KEY.encode())

    def tearDown(self):
        shutil.rmtree(self.test_dir)


class TestM2CryptoRegression47124(TestCase):

    SIGNATURE = (
        b"w\xac\xfe18o\xeb\xfb\x14+\x9e\xd1\xb7\x7fe}\xec\xd6\xe1P\x9e\xab"
        b"\xb5\x07\xe0\xc1\xfd\xda#\x04Z\x8d\x7f\x0b\x1f}:~\xb2s\x860u\x02N"
        b'\xd4q"\xb7\x86*\x8f\x1f\xd0\x9d\x11\x92\xc5~\xa68\xac>\x12H\xc2%y,'
        b"\xe6\xceU\x1e\xa3?\x0c,\xf0u\xbb\xd0[g_\xdd\x8b\xb0\x95:Y\x18\xa5*"
        b"\x99\xfd\xf3K\x92\x92 ({\xd1\xff\xd9F\xc8\xd6K\x86e\xf9\xa8\xad\xb0z"
        b"\xe3\x9dD\xf5k\x8b_<\xe7\xe7\xec\xf3\"'\xd5\xd2M\xb4\xce\x1a\xe3$"
        b"\x9c\x81\xad\xf9\x11\xf6\xf5>)\xc7\xdd\x03&\xf7\x86@ks\xa6\x05\xc2"
        b"\xd0\xbd\x1a7\xfc\xde\xe6\xb0\xad!\x12#\xc86Y\xea\xc5\xe3\xe2\xb3"
        b"\xc9\xaf\xfa\x0c\xf2?\xbf\x93w\x18\x9e\x0b\xa2a\x10:M\x05\x89\xe2W.Q"
        b"\xe8;yGT\xb1\xf2\xc6A\xd2\xc4\xbeN\xb3\xcfS\xaf\x03f\xe2\xb4)\xe7\xf6"
        b'\xdbs\xd0Z}8\xa4\xd2\x1fW*\xe6\x1c"\x8b\xd0\x18w\xb9\x7f\x9e\x96\xa3'
        b"\xd9v\xf7\x833\x8e\x01"
    )

    @pytest.mark.skipif(not HAS_M2, reason="Skip when m2crypto is not installed")
    def test_m2crypto_verify_bytes(self):
        message = salt.utils.stringutils.to_unicode("meh")
        with patch(
            "salt.utils.files.fopen",
            mock_open(read_data=salt.utils.stringutils.to_bytes(PUBKEY_DATA)),
        ):
            salt.crypt.verify_signature("/keydir/keyname.pub", message, self.SIGNATURE)

    @pytest.mark.skipif(not HAS_M2, reason="Skip when m2crypto is not installed")
    def test_m2crypto_verify_unicode(self):
        message = salt.utils.stringutils.to_bytes("meh")
        with patch(
            "salt.utils.files.fopen",
            mock_open(read_data=salt.utils.stringutils.to_bytes(PUBKEY_DATA)),
        ):
            salt.crypt.verify_signature("/keydir/keyname.pub", message, self.SIGNATURE)

    @pytest.mark.skipif(not HAS_M2, reason="Skip when m2crypto is not installed")
    def test_m2crypto_sign_bytes(self):
        message = salt.utils.stringutils.to_unicode("meh")
        key = M2Crypto.RSA.load_key_string(
            salt.utils.stringutils.to_bytes(PRIVKEY_DATA)
        )
        with patch("salt.crypt.get_rsa_key", return_value=key):
            signature = salt.crypt.sign_message(
                "/keydir/keyname.pem", message, passphrase="password"
            )
        self.assertEqual(signature, self.SIGNATURE)

    @pytest.mark.skipif(not HAS_M2, reason="Skip when m2crypto is not installed")
    def test_m2crypto_sign_unicode(self):
        message = salt.utils.stringutils.to_bytes("meh")
        key = M2Crypto.RSA.load_key_string(
            salt.utils.stringutils.to_bytes(PRIVKEY_DATA)
        )
        with patch("salt.crypt.get_rsa_key", return_value=key):
            signature = salt.crypt.sign_message(
                "/keydir/keyname.pem", message, passphrase="password"
            )
        self.assertEqual(signature, self.SIGNATURE)


@pytest.mark.skipif(
    not HAS_M2 and not HAS_PYCRYPTO_RSA,
    reason="No crypto library found. Install either M2Crypto or Cryptodome to run this test",
)
class TestCrypt(TestCase):
    def test_pwdata_decrypt(self):
        key_string = """-----BEGIN RSA PRIVATE KEY-----
MIIEpQIBAAKCAQEAzhBRyyHa7b63RLE71uKMKgrpulcAJjaIaN68ltXcCvy4w9pi
Kj+4I3Qp6RvUaHOEmymqyjOMjQc6iwpe0scCFqh3nUk5YYaLZ3WAW0htQVlnesgB
ZiBg9PBeTQY/LzqtudL6RCng/AX+fbnCsddlIysRxnUoNVMvz0gAmCY2mnTDjcTt
pyxuk2T0AHSHNCKCalm75L1bWDFF+UzFemf536tBfBUGRWR6jWTij85vvCntxHS/
HdknaTJ50E7XGVzwBJpCyV4Y2VXuW/3KrCNTqXw+jTmEw0vlcshfDg/vb3IxsUSK
5KuHalKq/nUIc+F4QCJOl+A10goGdIfYC1/67QIDAQABAoIBAAOP+qoFWtCTZH22
hq9PWVb8u0+yY1lFxhPyDdaZueUiu1r/coUCdv996Z+TEJgBr0AzdzVpsLtbbaKr
ujnwoNOdc/vvISPTfKN8P4zUcrcXgZd4z7VhR+vUH/0652q8m/ZDdHorMy2IOP8Z
cAk9DQ2PmA4TRm+tkX0G5KO8vWLsK921aRMWdsKJyQ0lYxl7M8JWupFsCJFr/U+8
dAVtwnUiS7RnhBABZ1cfNTHYhXVAh4d+a9y/gZ00a66OGqPxiXfhjjDUZ6fGvWKN
FlhKWEg6YqIx/H4aNXkLI5Rzzhdx/c2ukNm7+X2veRcAW7bcTwk8wxJxciEP5pBi
1el9VE0CgYEA/lbzdE2M4yRBvTfYYC6BqZcn+BqtrAUc2h3fEy+p7lwlet0af1id
gWpYpOJyLc0AUfR616/m2y3PwEH/nMKDSTuU7o/qKNtlHW0nQcnhDCjTUydS3+J/
JM3dhfgVqi03rjqNcgHA2eOEwcu/OBZtiaC0wqKbuRZRtfGffyoO3ssCgYEAz2iw
wqu/NkA+MdQIxz/a3Is7gGwoFu6h7O+XU2uN8Y2++jSBw9AzzWj31YCvyjuJPAE+
gxHm6yOnNoLVn423NtibHejhabzHNIK6UImH99bSTKabsxfF2BX6v982BimU1jwc
bYykzws37oN/poPb5FTpEiAUrsd2bAMn/1S43icCgYEAulHkY0z0aumCpyUkA8HO
BvjOtPiGRcAxFLBRXPLL3+vtIQachLHcIJRRf+jLkDXfiCo7W4pm6iWzTbqLkMEG
AD3/qowPFAM1Hct6uL01efzmYsIp+g0o60NMhvnolRQu+Bm4yM30AyqjdHzYBjSX
5fyuru8EeSCal1j8aOHcpuUCgYEAhGhDH6Pg59NPYSQJjpm3MMA59hwV473n5Yh2
xKyO6zwgRT6r8MPDrkhqnwQONT6Yt5PbwnT1Q/t4zhXsJnWkFwFk1U1MSeJYEa+7
HZsPECs2CfT6xPRSO0ac00y+AmUdPT8WruDwfbSdukh8f2MCR9vlBsswKPvxH7dM
G3aMplUCgYEAmMFgB/6Ox4OsQPPC6g4G+Ezytkc4iVkMEcjiVWzEsYATITjq3weO
/XDGBYJoBhYwWPi9oBufFc/2pNtWy1FKKXPuVyXQATdA0mfEPbtsHjMFQNZbeKnm
0na/SysSDCK3P+9ijlbjqLjMmPEmhJxGWTJ7khnTTkfre7/w9ZxJxi8=
-----END RSA PRIVATE KEY-----"""
        pwdata = b"""\
V\x80+b\xca\x06M\xb6\x12\xc6\xe8\xf2\xb5\xbb\xd8m\xc0\x97\x9a\xeb\xb9q\x19\xc3\
\xcdi\xb84\x90\xaf\x12kT\xe2@u\xd6\xe8T\x89\xa3\xc7\xb2Y\xd1N\x00\xa9\xc0"\xbe\
\xed\xb1\xc3\xb7^\xbf\xbd\x8b\x13\xd3/L\x1b\xa1`\xe2\xea\x03\x98\x82\xf3uS&|\
\xe5\xd8J\xce\xfc\x97\x8d\x0b\x949\xc0\xbd^\xef\xc6\xfd\xce\xbb\x1e\xd0"(m\xe1\
\x95\xfb\xc8/\x07\x93\xb8\xda\x8f\x99\xfe\xdc\xd5\xcb\xdb\xb2\xf11M\xdbD\xcf\
\x95\x13p\r\xa4\x1c{\xd5\xdb\xc7\xe5\xaf\x95F\x97\xa9\x00p~\xb5\xec\xa4.\xd0\
\xa4\xb4\xf4f\xcds,Y/\xa1:WF\xb8\xc7\x07\xaa\x0b<\'~\x1b$D9\xd4\x8d\xf0x\xc5\
\xee\xa8:\xe6\x00\x10\xc5i\x11\xc7]C8\x05l\x8b\x9b\xc3\x83e\xf7y\xadi:0\xb4R\
\x1a(\x04&yL8\x19s\n\x11\x81\xfd?\xfb2\x80Ll\xa1\xdc\xc9\xb6P\xca\x8d\'\x11\xc1\
\x07\xa5\xa1\x058\xc7\xce\xbeb\x92\xbf\x0bL\xec\xdf\xc3M\x83\xfb$\xec\xd5\xf9\
"""
        self.assertEqual("1234", salt.crypt.pwdata_decrypt(key_string, pwdata))



"""
Namecheap Domain Management

.. versionadded:: 2017.7.0

Prerequisites
-------------

This module uses the ``requests`` Python module to communicate to the namecheap
API.

Configuration
-------------

The Namecheap username, API key and URL should be set in the minion configuration
file, or in the Pillar data.

.. code-block:: yaml

    namecheap.name: companyname
    namecheap.key: a1b2c3d4e5f67a8b9c0d1e2f3
    namecheap.client_ip: 162.155.30.172
    #Real url
    namecheap.url: https://api.namecheap.com/xml.response
    #Sandbox url
    #namecheap.url: https://api.sandbox.namecheap.xml.response
"""

import logging

CAN_USE_NAMECHEAP = True

try:
    import salt.utils.namecheap
except ImportError:
    CAN_USE_NAMECHEAP = False


log = logging.getLogger(__name__)


def __virtual__():
    """
    Check to make sure requests and xml are installed and requests
    """
    if CAN_USE_NAMECHEAP:
        return "namecheap_domains"
    return False


def reactivate(domain_name):
    """
    Try to reactivate the expired domain name

    Returns the following information:

    - Whether or not the domain was reactivated successfully
    - The amount charged for reactivation
    - The order ID
    - The transaction ID

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_domains.reactivate my-domain-name
    """
    opts = salt.utils.namecheap.get_opts("namecheap.domains.reactivate")
    opts["DomainName"] = domain_name

    response_xml = salt.utils.namecheap.post_request(opts)

    if response_xml is None:
        return {}

    domainreactivateresult = response_xml.getElementsByTagName(
        "DomainReactivateResult"
    )[0]
    return salt.utils.namecheap.xml_to_dict(domainreactivateresult)


def renew(domain_name, years, promotion_code=None):
    """
    Try to renew the specified expiring domain name for a specified number of years

    domain_name
        The domain name to be renewed

    years
        Number of years to renew

    Returns the following information:

    - Whether or not the domain was renewed successfully
    - The domain ID
    - The order ID
    - The transaction ID
    - The amount charged for renewal

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_domains.renew my-domain-name 5
    """

    opts = salt.utils.namecheap.get_opts("namecheap.domains.renew")
    opts["DomainName"] = domain_name
    opts["Years"] = years
    if promotion_code is not None:
        opts["PromotionCode"] = promotion_code

    response_xml = salt.utils.namecheap.post_request(opts)

    if response_xml is None:
        return {}

    domainrenewresult = response_xml.getElementsByTagName("DomainRenewResult")[0]
    return salt.utils.namecheap.xml_to_dict(domainrenewresult)


def create(domain_name, years, **kwargs):
    """
    Try to register the specified domain name

    domain_name
        The domain name to be registered

    years
        Number of years to register

    Returns the following information:

    - Whether or not the domain was renewed successfully
    - Whether or not WhoisGuard is enabled
    - Whether or not registration is instant
    - The amount charged for registration
    - The domain ID
    - The order ID
    - The transaction ID

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_domains.create my-domain-name 2
    """
    idn_codes = (
        "afr",
        "alb",
        "ara",
        "arg",
        "arm",
        "asm",
        "ast",
        "ave",
        "awa",
        "aze",
        "bak",
        "bal",
        "ban",
        "baq",
        "bas",
        "bel",
        "ben",
        "bho",
        "bos",
        "bul",
        "bur",
        "car",
        "cat",
        "che",
        "chi",
        "chv",
        "cop",
        "cos",
        "cze",
        "dan",
        "div",
        "doi",
        "dut",
        "eng",
        "est",
        "fao",
        "fij",
        "fin",
        "fre",
        "fry",
        "geo",
        "ger",
        "gla",
        "gle",
        "gon",
        "gre",
        "guj",
        "heb",
        "hin",
        "hun",
        "inc",
        "ind",
        "inh",
        "isl",
        "ita",
        "jav",
        "jpn",
        "kas",
        "kaz",
        "khm",
        "kir",
        "kor",
        "kur",
        "lao",
        "lav",
        "lit",
        "ltz",
        "mal",
        "mkd",
        "mlt",
        "mol",
        "mon",
        "mri",
        "msa",
        "nep",
        "nor",
        "ori",
        "oss",
        "pan",
        "per",
        "pol",
        "por",
        "pus",
        "raj",
        "rum",
        "rus",
        "san",
        "scr",
        "sin",
        "slo",
        "slv",
        "smo",
        "snd",
        "som",
        "spa",
        "srd",
        "srp",
        "swa",
        "swe",
        "syr",
        "tam",
        "tel",
        "tgk",
        "tha",
        "tib",
        "tur",
        "ukr",
        "urd",
        "uzb",
        "vie",
        "wel",
        "yid",
    )

    require_opts = [
        "AdminAddress1",
        "AdminCity",
        "AdminCountry",
        "AdminEmailAddress",
        "AdminFirstName",
        "AdminLastName",
        "AdminPhone",
        "AdminPostalCode",
        "AdminStateProvince",
        "AuxBillingAddress1",
        "AuxBillingCity",
        "AuxBillingCountry",
        "AuxBillingEmailAddress",
        "AuxBillingFirstName",
        "AuxBillingLastName",
        "AuxBillingPhone",
        "AuxBillingPostalCode",
        "AuxBillingStateProvince",
        "RegistrantAddress1",
        "RegistrantCity",
        "RegistrantCountry",
        "RegistrantEmailAddress",
        "RegistrantFirstName",
        "RegistrantLastName",
        "RegistrantPhone",
        "RegistrantPostalCode",
        "RegistrantStateProvince",
        "TechAddress1",
        "TechCity",
        "TechCountry",
        "TechEmailAddress",
        "TechFirstName",
        "TechLastName",
        "TechPhone",
        "TechPostalCode",
        "TechStateProvince",
        "Years",
    ]
    opts = salt.utils.namecheap.get_opts("namecheap.domains.create")
    opts["DomainName"] = domain_name
    opts["Years"] = str(years)

    def add_to_opts(opts_dict, kwargs, value, suffix, prefices):
        for prefix in prefices:
            nextkey = prefix + suffix
            if nextkey not in kwargs:
                opts_dict[nextkey] = value

    for key, value in kwargs.items():
        if key.startswith("Registrant"):
            add_to_opts(
                opts,
                kwargs,
                value,
                key[10:],
                ["Tech", "Admin", "AuxBilling", "Billing"],
            )

        if key.startswith("Tech"):
            add_to_opts(
                opts,
                kwargs,
                value,
                key[4:],
                ["Registrant", "Admin", "AuxBilling", "Billing"],
            )

        if key.startswith("Admin"):
            add_to_opts(
                opts,
                kwargs,
                value,
                key[5:],
                ["Registrant", "Tech", "AuxBilling", "Billing"],
            )

        if key.startswith("AuxBilling"):
            add_to_opts(
                opts,
                kwargs,
                value,
                key[10:],
                ["Registrant", "Tech", "Admin", "Billing"],
            )

        if key.startswith("Billing"):
            add_to_opts(
                opts,
                kwargs,
                value,
                key[7:],
                ["Registrant", "Tech", "Admin", "AuxBilling"],
            )

        if key == "IdnCode" and key not in idn_codes:
            log.error("Invalid IdnCode")
            raise Exception("Invalid IdnCode")

        opts[key] = value

    for requiredkey in require_opts:
        if requiredkey not in opts:
            log.error("Missing required parameter '%s'", requiredkey)
            raise Exception(f"Missing required parameter '{requiredkey}'")

    response_xml = salt.utils.namecheap.post_request(opts)

    if response_xml is None:
        return {}

    domainresult = response_xml.getElementsByTagName("DomainCreateResult")[0]
    return salt.utils.namecheap.atts_to_dict(domainresult)


def check(*domains_to_check):
    """
    Checks the availability of domains

    domains_to_check
        array of strings  List of domains to check

    Returns a dictionary mapping the each domain name to a boolean denoting
    whether or not it is available.

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_domains.check domain-to-check
    """
    opts = salt.utils.namecheap.get_opts("namecheap.domains.check")
    opts["DomainList"] = ",".join(domains_to_check)

    response_xml = salt.utils.namecheap.get_request(opts)

    if response_xml is None:
        return {}

    domains_checked = {}
    for result in response_xml.getElementsByTagName("DomainCheckResult"):
        available = result.getAttribute("Available")
        domains_checked[result.getAttribute("Domain").lower()] = (
            salt.utils.namecheap.string_to_value(available)
        )

    return domains_checked


def get_info(domain_name):
    """
    Returns information about the requested domain

    returns a dictionary of information about the domain_name

    domain_name
        string  Domain name to get information about

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_domains.get_info my-domain-name
    """
    opts = salt.utils.namecheap.get_opts("namecheap.domains.getinfo")
    opts["DomainName"] = domain_name

    response_xml = salt.utils.namecheap.get_request(opts)

    if response_xml is None:
        return []

    domaingetinforesult = response_xml.getElementsByTagName("DomainGetInfoResult")[0]

    return salt.utils.namecheap.xml_to_dict(domaingetinforesult)


def get_tld_list():
    """
    Returns a list of TLDs as objects

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_domains.get_tld_list
    """

    response_xml = salt.utils.namecheap.get_request(
        salt.utils.namecheap.get_opts("namecheap.domains.gettldlist")
    )

    if response_xml is None:
        return []

    tldresult = response_xml.getElementsByTagName("Tlds")[0]
    tlds = []

    for e in tldresult.getElementsByTagName("Tld"):
        tld = salt.utils.namecheap.atts_to_dict(e)
        tld["data"] = e.firstChild.data
        categories = []
        subcategories = e.getElementsByTagName("Categories")[0]
        for c in subcategories.getElementsByTagName("TldCategory"):
            categories.append(salt.utils.namecheap.atts_to_dict(c))
        tld["categories"] = categories
        tlds.append(tld)

    return tlds


def get_list(list_type=None, search_term=None, page=None, page_size=None, sort_by=None):
    """
    Returns a list of domains for the particular user as a list of objects
    offset by ``page`` length of ``page_size``

    list_type : ALL
        One of ``ALL``, ``EXPIRING``, ``EXPIRED``

    search_term
        Keyword to look for on the domain list

    page : 1
        Number of result page to return

    page_size : 20
        Number of domains to be listed per page (minimum: ``10``, maximum:
        ``100``)

    sort_by
        One of ``NAME``, ``NAME_DESC``, ``EXPIREDATE``, ``EXPIREDATE_DESC``,
        ``CREATEDATE``, or ``CREATEDATE_DESC``

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_domains.get_list
    """
    opts = salt.utils.namecheap.get_opts("namecheap.domains.getList")

    if list_type is not None:
        if list_type not in ["ALL", "EXPIRING", "EXPIRED"]:
            log.error("Invalid option for list_type")
            raise Exception("Invalid option for list_type")
        opts["ListType"] = list_type

    if search_term is not None:
        if len(search_term) > 70:
            log.warning("search_term trimmed to first 70 characters")
            search_term = search_term[0:70]
        opts["SearchTerm"] = search_term

    if page is not None:
        opts["Page"] = page

    if page_size is not None:
        if page_size > 100 or page_size < 10:
            log.error("Invalid option for page")
            raise Exception("Invalid option for page")
        opts["PageSize"] = page_size

    if sort_by is not None:
        if sort_by not in [
            "NAME",
            "NAME_DESC",
            "EXPIREDATE",
            "EXPIREDATE_DESC",
            "CREATEDATE",
            "CREATEDATE_DESC",
        ]:
            log.error("Invalid option for sort_by")
            raise Exception("Invalid option for sort_by")
        opts["SortBy"] = sort_by

    response_xml = salt.utils.namecheap.get_request(opts)

    if response_xml is None:
        return []

    domainresult = response_xml.getElementsByTagName("DomainGetListResult")[0]

    domains = []
    for d in domainresult.getElementsByTagName("Domain"):
        domains.append(salt.utils.namecheap.atts_to_dict(d))

    return domains



"""
1&1 Cloud Server Module
=======================

The 1&1 SaltStack cloud module allows a 1&1 server to be automatically deployed
and bootstrapped with Salt. It also has functions to create block storages and
ssh keys.

:depends: 1and1 >= 1.2.0

The module requires the 1&1 api_token to be provided.  The server should also
be assigned a public LAN, a private LAN, or both along with SSH key pairs.

Set up the cloud configuration at ``/etc/salt/cloud.providers`` or
``/etc/salt/cloud.providers.d/oneandone.conf``:

.. code-block:: yaml

    my-oneandone-config:
      driver: oneandone
      # The 1&1 api token
      api_token: <your-token>
      # SSH private key filename
      ssh_private_key: /path/to/private_key
      # SSH public key filename
      ssh_public_key: /path/to/public_key

.. code-block:: yaml

    my-oneandone-profile:
      provider: my-oneandone-config
      # Either provide fixed_instance_size_id or vcore, cores_per_processor, ram, and hdds.
      # Size of the ID desired for the server
      fixed_instance_size: S
      # Total amount of processors
      vcore: 2
      # Number of cores per processor
      cores_per_processor: 2
      # RAM memory size in GB
      ram: 4
      # Hard disks
      hdds:
      -
        is_main: true
        size: 20
      -
        is_main: false
        size: 20
      # ID of the appliance image that will be installed on server
      appliance_id: <ID>
      # ID of the datacenter where the server will be created
      datacenter_id: <ID>
      # Description of the server
      description: My server description
      # Password of the server. Password must contain more than 8 characters
      # using uppercase letters, numbers and other special symbols.
      password: P4$$w0rD
      # Power on server after creation - default True
      power_on: true
      # Firewall policy ID. If it is not provided, the server will assign
      # the best firewall policy, creating a new one if necessary.
      # If the parameter is sent with a 0 value, the server will be created with all ports blocked.
      firewall_policy_id: <ID>
      # IP address ID
      ip_id: <ID>
      # Load balancer ID
      load_balancer_id: <ID>
      # Monitoring policy ID
      monitoring_policy_id: <ID>

Set ``deploy`` to False if Salt should not be installed on the node.

.. code-block:: yaml

    my-oneandone-profile:
      deploy: False

Create an SSH key

.. code-block:: bash

    sudo salt-cloud -f create_ssh_key my-oneandone-config name='SaltTest' description='SaltTestDescription'

Create a block storage

.. code-block:: bash

    sudo salt-cloud -f create_block_storage my-oneandone-config name='SaltTest2'
    description='SaltTestDescription' size=50 datacenter_id='5091F6D8CBFEF9C26ACE957C652D5D49'

"""

import logging
import os
import pprint
import time

import salt.config as config
import salt.utils.cloud
import salt.utils.files
import salt.utils.stringutils
from salt.exceptions import (
    SaltCloudConfigError,
    SaltCloudExecutionFailure,
    SaltCloudExecutionTimeout,
    SaltCloudNotFound,
    SaltCloudSystemExit,
)

try:
    # pylint: disable=no-name-in-module
    from oneandone.client import BlockStorage, Hdd, OneAndOneService, Server, SshKey

    # pylint: enable=no-name-in-module

    HAS_ONEANDONE = True
except ImportError:
    HAS_ONEANDONE = False

# Get logging started
log = logging.getLogger(__name__)

__virtualname__ = "oneandone"


# Only load in this module if the 1&1 configurations are in place
def __virtual__():
    """
    Check for 1&1 configurations.
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__, _get_active_provider_name() or __virtualname__, ("api_token",)
    )


def get_dependencies():
    """
    Warn if dependencies are not met.
    """
    return config.check_driver_dependencies(
        __virtualname__, {"oneandone": HAS_ONEANDONE}
    )


def get_conn():
    """
    Return a conn object for the passed VM data
    """
    return OneAndOneService(
        api_token=config.get_cloud_config_value(
            "api_token", get_configured_provider(), __opts__, search_global=False
        )
    )


def get_size(vm_):
    """
    Return the VM's size object
    """
    vm_size = config.get_cloud_config_value(
        "fixed_instance_size", vm_, __opts__, default=None, search_global=False
    )
    sizes = avail_sizes()

    if not vm_size:
        size = next((item for item in sizes if item["name"] == "S"), None)
        return size

    size = next(
        (item for item in sizes if item["name"] == vm_size or item["id"] == vm_size),
        None,
    )
    if size:
        return size

    raise SaltCloudNotFound(f"The specified size, '{vm_size}', could not be found.")


def get_image(vm_):
    """
    Return the image object to use
    """
    vm_image = config.get_cloud_config_value("image", vm_, __opts__).encode(
        "ascii", "salt-cloud-force-ascii"
    )

    images = avail_images()
    for key, value in images.items():
        if vm_image and vm_image in (images[key]["id"], images[key]["name"]):
            return images[key]

    raise SaltCloudNotFound(f"The specified image, '{vm_image}', could not be found.")


def avail_locations(conn=None, call=None):
    """
    List available locations/datacenters for 1&1
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    datacenters = []

    if not conn:
        conn = get_conn()

    for datacenter in conn.list_datacenters():
        datacenters.append({datacenter["country_code"]: datacenter})

    return {"Locations": datacenters}


def create_block_storage(kwargs=None, call=None):
    """
    Create a block storage
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    conn = get_conn()

    # Assemble the composite block storage object.
    block_storage = _get_block_storage(kwargs)

    data = conn.create_block_storage(block_storage=block_storage)

    return {"BlockStorage": data}


def _get_block_storage(kwargs):
    """
    Construct a block storage instance from passed arguments
    """
    if kwargs is None:
        kwargs = {}

    block_storage_name = kwargs.get("name", None)
    block_storage_size = kwargs.get("size", None)
    block_storage_description = kwargs.get("description", None)
    datacenter_id = kwargs.get("datacenter_id", None)
    server_id = kwargs.get("server_id", None)

    block_storage = BlockStorage(name=block_storage_name, size=block_storage_size)

    if block_storage_description:
        block_storage.description = block_storage_description

    if datacenter_id:
        block_storage.datacenter_id = datacenter_id

    if server_id:
        block_storage.server_id = server_id

    return block_storage


def _get_ssh_key(kwargs):
    """
    Construct an SshKey instance from passed arguments
    """
    ssh_key_name = kwargs.get("name", None)
    ssh_key_description = kwargs.get("description", None)
    public_key = kwargs.get("public_key", None)

    return SshKey(
        name=ssh_key_name, description=ssh_key_description, public_key=public_key
    )


def create_ssh_key(kwargs=None, call=None):
    """
    Create an ssh key
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    conn = get_conn()

    # Assemble the composite SshKey object.
    ssh_key = _get_ssh_key(kwargs)

    data = conn.create_ssh_key(ssh_key=ssh_key)

    return {"SshKey": data}


def avail_images(conn=None, call=None):
    """
    Return a list of the server appliances that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-images option"
        )

    if not conn:
        conn = get_conn()

    ret = {}

    for appliance in conn.list_appliances():
        ret[appliance["name"]] = appliance

    return ret


def avail_sizes(call=None):
    """
    Return a dict of all available VM sizes on the cloud provider with
    relevant data.
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with "
            "-f or --function, or with the --list-sizes option"
        )

    conn = get_conn()

    sizes = conn.fixed_server_flavors()

    return sizes


def script(vm_):
    """
    Return the script deployment object
    """
    return salt.utils.cloud.os_script(
        config.get_cloud_config_value("script", vm_, __opts__),
        vm_,
        __opts__,
        salt.utils.cloud.salt_config_to_yaml(
            salt.utils.cloud.minion_config(__opts__, vm_)
        ),
    )


def list_nodes(conn=None, call=None):
    """
    Return a list of VMs that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )

    if not conn:
        conn = get_conn()

    ret = {}
    nodes = conn.list_servers()

    for node in nodes:
        public_ips = []
        private_ips = []
        ret = {}

        size = node.get("hardware").get("fixed_instance_size_id", "Custom size")

        if node.get("private_networks"):
            for private_ip in node["private_networks"]:
                private_ips.append(private_ip)

        if node.get("ips"):
            for public_ip in node["ips"]:
                public_ips.append(public_ip["ip"])

        server = {
            "id": node["id"],
            "image": node["image"]["id"],
            "size": size,
            "state": node["status"]["state"],
            "private_ips": private_ips,
            "public_ips": public_ips,
        }
        ret[node["name"]] = server

    return ret


def list_nodes_full(conn=None, call=None):
    """
    Return a list of the VMs that are on the provider, with all fields
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function."
        )

    if not conn:
        conn = get_conn()

    ret = {}
    nodes = conn.list_servers()

    for node in nodes:
        ret[node["name"]] = node

    return ret


def list_nodes_select(conn=None, call=None):
    """
    Return a list of the VMs that are on the provider, with select fields
    """
    if not conn:
        conn = get_conn()

    return salt.utils.cloud.list_nodes_select(
        list_nodes_full(conn, "function"),
        __opts__["query.selection"],
        call,
    )


def show_instance(name, call=None):
    """
    Show the details from the provider concerning an instance
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )

    nodes = list_nodes_full()
    __utils__["cloud.cache_node"](nodes[name], _get_active_provider_name(), __opts__)
    return nodes[name]


def _get_server(vm_):
    """
    Construct server instance from cloud profile config
    """
    description = config.get_cloud_config_value(
        "description", vm_, __opts__, default=None, search_global=False
    )

    ssh_key = load_public_key(vm_)

    vcore = None
    cores_per_processor = None
    ram = None
    fixed_instance_size_id = None

    if "fixed_instance_size" in vm_:
        fixed_instance_size = get_size(vm_)
        fixed_instance_size_id = fixed_instance_size["id"]
    elif vm_["vcore"] and vm_["cores_per_processor"] and vm_["ram"] and vm_["hdds"]:
        vcore = config.get_cloud_config_value(
            "vcore", vm_, __opts__, default=None, search_global=False
        )
        cores_per_processor = config.get_cloud_config_value(
            "cores_per_processor", vm_, __opts__, default=None, search_global=False
        )
        ram = config.get_cloud_config_value(
            "ram", vm_, __opts__, default=None, search_global=False
        )
    else:
        raise SaltCloudConfigError(
            "'fixed_instance_size' or 'vcore',"
            "'cores_per_processor', 'ram', and 'hdds'"
            "must be provided."
        )

    appliance_id = config.get_cloud_config_value(
        "appliance_id", vm_, __opts__, default=None, search_global=False
    )

    password = config.get_cloud_config_value(
        "password", vm_, __opts__, default=None, search_global=False
    )

    firewall_policy_id = config.get_cloud_config_value(
        "firewall_policy_id", vm_, __opts__, default=None, search_global=False
    )

    ip_id = config.get_cloud_config_value(
        "ip_id", vm_, __opts__, default=None, search_global=False
    )

    load_balancer_id = config.get_cloud_config_value(
        "load_balancer_id", vm_, __opts__, default=None, search_global=False
    )

    monitoring_policy_id = config.get_cloud_config_value(
        "monitoring_policy_id", vm_, __opts__, default=None, search_global=False
    )

    datacenter_id = config.get_cloud_config_value(
        "datacenter_id", vm_, __opts__, default=None, search_global=False
    )

    private_network_id = config.get_cloud_config_value(
        "private_network_id", vm_, __opts__, default=None, search_global=False
    )

    power_on = config.get_cloud_config_value(
        "power_on", vm_, __opts__, default=True, search_global=False
    )

    public_key = config.get_cloud_config_value(
        "public_key_ids", vm_, __opts__, default=True, search_global=False
    )

    # Contruct server object
    return Server(
        name=vm_["name"],
        description=description,
        fixed_instance_size_id=fixed_instance_size_id,
        vcore=vcore,
        cores_per_processor=cores_per_processor,
        ram=ram,
        appliance_id=appliance_id,
        password=password,
        power_on=power_on,
        firewall_policy_id=firewall_policy_id,
        ip_id=ip_id,
        load_balancer_id=load_balancer_id,
        monitoring_policy_id=monitoring_policy_id,
        datacenter_id=datacenter_id,
        rsa_key=ssh_key,
        private_network_id=private_network_id,
        public_key=public_key,
    )


def _get_hdds(vm_):
    """
    Construct VM hdds from cloud profile config
    """
    _hdds = config.get_cloud_config_value(
        "hdds", vm_, __opts__, default=None, search_global=False
    )

    hdds = []

    for hdd in _hdds:
        hdds.append(Hdd(size=hdd["size"], is_main=hdd["is_main"]))

    return hdds


def create(vm_):
    """
    Create a single VM from a data dict
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__, (_get_active_provider_name() or "oneandone"), vm_["profile"]
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    data = None
    conn = get_conn()
    hdds = []

    # Assemble the composite server object.
    server = _get_server(vm_)

    if not bool(server.specs["hardware"]["fixed_instance_size_id"]):
        # Assemble the hdds object.
        hdds = _get_hdds(vm_)

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args={"name": vm_["name"]},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    try:
        data = conn.create_server(server=server, hdds=hdds)

        _wait_for_completion(conn, get_wait_timeout(vm_), data["id"])
    except Exception as exc:  # pylint: disable=W0703
        log.error(
            "Error creating %s on 1and1\n\n"
            "The following exception was thrown by the 1and1 library "
            "when trying to run the initial deployment: \n%s",
            vm_["name"],
            exc,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False

    vm_["server_id"] = data["id"]
    password = data["first_password"]

    def __query_node_data(vm_, data):
        """
        Query node data until node becomes available.
        """
        running = False
        try:
            data = show_instance(vm_["name"], "action")
            if not data:
                return False
            log.debug(
                "Loaded node data for %s:\nname: %s\nstate: %s",
                vm_["name"],
                pprint.pformat(data["name"]),
                data["status"]["state"],
            )
        except Exception as err:  # pylint: disable=broad-except
            log.error(
                "Failed to get nodes list: %s",
                err,
                # Show the trackback if the debug logging level is enabled
                exc_info_on_loglevel=logging.DEBUG,
            )
            # Trigger a failure in the wait for IP function
            return False

        running = data["status"]["state"].lower() == "powered_on"
        if not running:
            # Still not running, trigger another iteration
            return

        vm_["ssh_host"] = data["ips"][0]["ip"]

        return data

    try:
        data = salt.utils.cloud.wait_for_ip(
            __query_node_data,
            update_args=(vm_, data),
            timeout=config.get_cloud_config_value(
                "wait_for_ip_timeout", vm_, __opts__, default=10 * 60
            ),
            interval=config.get_cloud_config_value(
                "wait_for_ip_interval", vm_, __opts__, default=10
            ),
        )
    except (SaltCloudExecutionTimeout, SaltCloudExecutionFailure) as exc:
        try:
            # It might be already up, let's destroy it!
            destroy(vm_["name"])
        except SaltCloudSystemExit:
            pass
        finally:
            raise SaltCloudSystemExit(str(exc.message))

    log.debug("VM is now running")
    log.info("Created Cloud VM %s", vm_)
    log.debug("%s VM creation details:\n%s", vm_, pprint.pformat(data))

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args={
            "name": vm_["name"],
            "profile": vm_["profile"],
            "provider": vm_["driver"],
        },
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    if "ssh_host" in vm_:
        vm_["password"] = password
        vm_["key_filename"] = get_key_filename(vm_)
        ret = __utils__["cloud.bootstrap"](vm_, __opts__)
        ret.update(data)
        return ret
    else:
        raise SaltCloudSystemExit("A valid IP address was not found.")


def destroy(name, call=None):
    """
    destroy a server by name

    :param name: name given to the server
    :param call: call value in this case is 'action'
    :return: array of booleans , true if successfully stopped and true if
             successfully removed

    CLI Example:

    .. code-block:: bash

        salt-cloud -d vm_name

    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        f"salt/cloud/{name}/destroying",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    conn = get_conn()
    node = get_node(conn, name)

    conn.delete_server(server_id=node["id"])

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        f"salt/cloud/{name}/destroyed",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    if __opts__.get("update_cachedir", False) is True:
        __utils__["cloud.delete_minion_cachedir"](
            name, _get_active_provider_name().split(":")[0], __opts__
        )

    return True


def reboot(name, call=None):
    """
    reboot a server by name
    :param name: name given to the machine
    :param call: call value in this case is 'action'
    :return: true if successful

    CLI Example:

    .. code-block:: bash

        salt-cloud -a reboot vm_name
    """
    conn = get_conn()
    node = get_node(conn, name)

    conn.modify_server_status(server_id=node["id"], action="REBOOT")

    return True


def stop(name, call=None):
    """
    stop a server by name
    :param name: name given to the machine
    :param call: call value in this case is 'action'
    :return: true if successful

    CLI Example:

    .. code-block:: bash

        salt-cloud -a stop vm_name
    """
    conn = get_conn()
    node = get_node(conn, name)

    conn.stop_server(server_id=node["id"])

    return True


def start(name, call=None):
    """
    start a server by name
    :param name: name given to the machine
    :param call: call value in this case is 'action'
    :return: true if successful

    CLI Example:

    .. code-block:: bash

        salt-cloud -a start vm_name
    """
    conn = get_conn()
    node = get_node(conn, name)

    conn.start_server(server_id=node["id"])

    return True


def get_node(conn, name):
    """
    Return a node for the named VM
    """
    for node in conn.list_servers(per_page=1000):
        if node["name"] == name:
            return node


def get_key_filename(vm_):
    """
    Check SSH private key file and return absolute path if exists.
    """
    key_filename = config.get_cloud_config_value(
        "ssh_private_key", vm_, __opts__, search_global=False, default=None
    )
    if key_filename is not None:
        key_filename = os.path.expanduser(key_filename)
        if not os.path.isfile(key_filename):
            raise SaltCloudConfigError(
                f"The defined ssh_private_key '{key_filename}' does not exist"
            )

        return key_filename


def load_public_key(vm_):
    """
    Load the public key file if exists.
    """
    public_key_filename = config.get_cloud_config_value(
        "ssh_public_key", vm_, __opts__, search_global=False, default=None
    )
    if public_key_filename is not None:
        public_key_filename = os.path.expanduser(public_key_filename)
        if not os.path.isfile(public_key_filename):
            raise SaltCloudConfigError(
                "The defined ssh_public_key '{}' does not exist".format(
                    public_key_filename
                )
            )

        with salt.utils.files.fopen(public_key_filename, "r") as public_key:
            key = salt.utils.stringutils.to_unicode(public_key.read().replace("\n", ""))

            return key


def get_wait_timeout(vm_):
    """
    Return the wait_for_timeout for resource provisioning.
    """
    return config.get_cloud_config_value(
        "wait_for_timeout", vm_, __opts__, default=15 * 60, search_global=False
    )


def _wait_for_completion(conn, wait_timeout, server_id):
    """
    Poll request status until resource is provisioned.
    """
    wait_timeout = time.time() + wait_timeout
    while wait_timeout > time.time():
        time.sleep(5)

        server = conn.get_server(server_id)
        server_state = server["status"]["state"].lower()

        if server_state == "powered_on":
            return
        elif server_state == "failed":
            raise Exception(f"Server creation failed for {server_id}")
        elif server_state in ("active", "enabled", "deploying", "configuring"):
            continue
        else:
            raise Exception(f"Unknown server state {server_state}")
    raise Exception(f"Timed out waiting for server create completion for {server_id}")



"""
    :codeauthor: :email:`Alexandru Bleotu <alexandru.bleotu@morganstanley.com>`

    Tests for esxdatacenter proxy
"""

import pytest

import salt.exceptions
import salt.proxy.esxdatacenter as esxdatacenter
from salt.config.schemas.esxdatacenter import EsxdatacenterProxySchema
from tests.support.mock import MagicMock, patch

try:
    import jsonschema

    HAS_JSONSCHEMA = True
except ImportError:
    HAS_JSONSCHEMA = False


pytestmark = [pytest.mark.skipif(not HAS_JSONSCHEMA, reason="jsonschema is required")]


@pytest.fixture
def opts_userpass():
    return {
        "proxy": {
            "proxytype": "esxdatacenter",
            "vcenter": "fake_vcenter",
            "datacenter": "fake_dc",
            "mechanism": "userpass",
            "username": "fake_username",
            "passwords": ["fake_password"],
            "protocol": "fake_protocol",
            "port": 100,
        }
    }


@pytest.fixture
def opts_sspi():
    return {
        "proxy": {
            "proxytype": "esxdatacenter",
            "vcenter": "fake_vcenter",
            "datacenter": "fake_dc",
            "mechanism": "sspi",
            "domain": "fake_domain",
            "principal": "fake_principal",
            "protocol": "fake_protocol",
            "port": 100,
        }
    }


@pytest.fixture
def configure_loader_modules(opts_sspi):
    with patch.dict(esxdatacenter.DETAILS):
        with patch(
            "salt.proxy.esxdatacenter.merge",
            MagicMock(return_value=opts_sspi["proxy"]),
        ):
            yield {esxdatacenter: {"__pillar__": {}}}


def test_merge(opts_sspi):
    mock_pillar_proxy = MagicMock()
    mock_opts_proxy = MagicMock()
    mock_merge = MagicMock(return_value=opts_sspi["proxy"])
    with patch.dict(esxdatacenter.__pillar__, {"proxy": mock_pillar_proxy}):
        with patch("salt.proxy.esxdatacenter.merge", mock_merge):
            esxdatacenter.init(opts={"proxy": mock_opts_proxy})
    mock_merge.assert_called_once_with(mock_opts_proxy, mock_pillar_proxy)


def test_esxdatacenter_schema(opts_sspi):
    mock_json_validate = MagicMock()
    serialized_schema = EsxdatacenterProxySchema().serialize()
    with patch("salt.proxy.esxdatacenter.jsonschema.validate", mock_json_validate):
        esxdatacenter.init(opts_sspi)
    mock_json_validate.assert_called_once_with(opts_sspi["proxy"], serialized_schema)


def test_invalid_proxy_input_error(opts_userpass):
    with patch(
        "salt.proxy.esxdatacenter.jsonschema.validate",
        MagicMock(
            side_effect=jsonschema.exceptions.ValidationError("Validation Error")
        ),
    ):
        with pytest.raises(salt.exceptions.InvalidConfigError) as excinfo:
            esxdatacenter.init(opts_userpass)
    assert excinfo.value.message == "Validation Error"


def test_no_username(opts_userpass):
    opts = opts_userpass.copy()
    del opts["proxy"]["username"]
    with patch("salt.proxy.esxdatacenter.merge", MagicMock(return_value=opts["proxy"])):
        with pytest.raises(salt.exceptions.InvalidConfigError) as excinfo:
            esxdatacenter.init(opts)
    assert (
        excinfo.value.message == "Mechanism is set to 'userpass', but no "
        "'username' key found in proxy config."
    )


def test_no_passwords(opts_userpass):
    opts = opts_userpass.copy()
    del opts["proxy"]["passwords"]
    with patch("salt.proxy.esxdatacenter.merge", MagicMock(return_value=opts["proxy"])):
        with pytest.raises(salt.exceptions.InvalidConfigError) as excinfo:
            esxdatacenter.init(opts)
    assert (
        excinfo.value.message == "Mechanism is set to 'userpass', but no "
        "'passwords' key found in proxy config."
    )


def test_no_domain(opts_sspi):
    opts = opts_sspi.copy()
    del opts["proxy"]["domain"]
    with patch("salt.proxy.esxdatacenter.merge", MagicMock(return_value=opts["proxy"])):
        with pytest.raises(salt.exceptions.InvalidConfigError) as excinfo:
            esxdatacenter.init(opts)
    assert (
        excinfo.value.message
        == "Mechanism is set to 'sspi', but no 'domain' key found in proxy config."
    )


def test_no_principal(opts_sspi):
    opts = opts_sspi.copy()
    del opts["proxy"]["principal"]
    with patch("salt.proxy.esxdatacenter.merge", MagicMock(return_value=opts["proxy"])):
        with pytest.raises(salt.exceptions.InvalidConfigError) as excinfo:
            esxdatacenter.init(opts)
    assert (
        excinfo.value.message
        == "Mechanism is set to 'sspi', but no 'principal' key found in proxy config."
    )


def test_find_credentials(opts_userpass):
    mock_find_credentials = MagicMock(return_value=("fake_username", "fake_password"))
    with patch(
        "salt.proxy.esxdatacenter.merge",
        MagicMock(return_value=opts_userpass["proxy"]),
    ):
        with patch("salt.proxy.esxdatacenter.find_credentials", mock_find_credentials):
            esxdatacenter.init(opts_userpass)
    mock_find_credentials.assert_called_once_with()


def test_details_userpass(opts_userpass):
    mock_find_credentials = MagicMock(return_value=("fake_username", "fake_password"))
    with patch(
        "salt.proxy.esxdatacenter.merge",
        MagicMock(return_value=opts_userpass["proxy"]),
    ):
        with patch("salt.proxy.esxdatacenter.find_credentials", mock_find_credentials):
            esxdatacenter.init(opts_userpass)
    assert esxdatacenter.DETAILS == {
        "vcenter": "fake_vcenter",
        "datacenter": "fake_dc",
        "mechanism": "userpass",
        "username": "fake_username",
        "password": "fake_password",
        "passwords": ["fake_password"],
        "protocol": "fake_protocol",
        "port": 100,
    }


def test_details_sspi(opts_sspi):
    esxdatacenter.init(opts_sspi)
    assert esxdatacenter.DETAILS == {
        "vcenter": "fake_vcenter",
        "datacenter": "fake_dc",
        "mechanism": "sspi",
        "domain": "fake_domain",
        "principal": "fake_principal",
        "protocol": "fake_protocol",
        "port": 100,
    }



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    TestCase for the salt.modules.linux_lvm module
"""

import os.path

import pytest

import salt.modules.linux_lvm as linux_lvm
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {linux_lvm: {}}


def test_version():
    """
    Tests LVM version info from lvm version
    """
    mock = MagicMock(
        return_value=(
            "  LVM version:     2.02.168(2) (2016-11-30)\n"
            "  Library version: 1.03.01 (2016-11-30)\n"
            "  Driver version:  4.35.0\n"
        )
    )
    with patch.dict(linux_lvm.__salt__, {"cmd.run": mock}):
        assert linux_lvm.version() == "2.02.168(2) (2016-11-30)"


def test_fullversion():
    """
    Tests all version info from lvm version
    """
    mock = MagicMock(
        return_value=(
            "  LVM version:     2.02.168(2) (2016-11-30)\n"
            "  Library version: 1.03.01 (2016-11-30)\n"
            "  Driver version:  4.35.0\n"
        )
    )
    with patch.dict(linux_lvm.__salt__, {"cmd.run": mock}):
        assert linux_lvm.fullversion() == {
            "LVM version": "2.02.168(2) (2016-11-30)",
            "Library version": "1.03.01 (2016-11-30)",
            "Driver version": "4.35.0",
        }


def test_pvdisplay():
    """
    Tests information about the physical volume(s)
    """
    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.pvdisplay() == {}

    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.pvdisplay(quiet=True) == {}
        mock.assert_called_with(
            ["pvdisplay", "-c"], ignore_retcode=True, python_shell=False
        )

    mock = MagicMock(return_value={"retcode": 0, "stdout": "A:B:C:D:E:F:G:H:I:J:K"})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.pvdisplay() == {
            "A": {
                "Allocated Physical Extents": "K",
                "Current Logical Volumes Here": "G",
                "Free Physical Extents": "J",
                "Internal Physical Volume Number": "D",
                "Physical Extent Size (kB)": "H",
                "Physical Volume (not) Allocatable": "F",
                "Physical Volume Device": "A",
                "Physical Volume Size (kB)": "C",
                "Physical Volume Status": "E",
                "Total Physical Extents": "I",
                "Volume Group Name": "B",
            }
        }

        mockpath = MagicMock(return_value="Z")
        with patch.object(os.path, "realpath", mockpath):
            assert linux_lvm.pvdisplay(real=True) == {
                "Z": {
                    "Allocated Physical Extents": "K",
                    "Current Logical Volumes Here": "G",
                    "Free Physical Extents": "J",
                    "Internal Physical Volume Number": "D",
                    "Physical Extent Size (kB)": "H",
                    "Physical Volume (not) Allocatable": "F",
                    "Physical Volume Device": "A",
                    "Physical Volume Size (kB)": "C",
                    "Physical Volume Status": "E",
                    "Real Physical Volume Device": "Z",
                    "Total Physical Extents": "I",
                    "Volume Group Name": "B",
                }
            }


def test_vgdisplay():
    """
    Tests information about the volume group(s)
    """
    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.vgdisplay() == {}

    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.vgdisplay(quiet=True) == {}
        mock.assert_called_with(
            ["vgdisplay", "-c"], ignore_retcode=True, python_shell=False
        )

    mock = MagicMock(
        return_value={"retcode": 0, "stdout": "A:B:C:D:E:F:G:H:I:J:K:L:M:N:O:P:Q"}
    )
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.vgdisplay() == {
            "A": {
                "Actual Physical Volumes": "K",
                "Allocated Physical Extents": "O",
                "Current Logical Volumes": "F",
                "Current Physical Volumes": "J",
                "Free Physical Extents": "P",
                "Internal Volume Group Number": "D",
                "Maximum Logical Volume Size": "H",
                "Maximum Logical Volumes": "E",
                "Maximum Physical Volumes": "I",
                "Open Logical Volumes": "G",
                "Physical Extent Size (kB)": "M",
                "Total Physical Extents": "N",
                "UUID": "Q",
                "Volume Group Access": "B",
                "Volume Group Name": "A",
                "Volume Group Size (kB)": "L",
                "Volume Group Status": "C",
            }
        }


def test_lvdisplay():
    """
    Return information about the logical volume(s)
    """
    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.lvdisplay() == {}

    mock = MagicMock(return_value={"retcode": 0, "stdout": "A:B:C:D:E:F:G:H:I:J:K:L:M"})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.lvdisplay() == {
            "A": {
                "Allocated Logical Extents": "I",
                "Allocation Policy": "J",
                "Current Logical Extents Associated": "H",
                "Internal Logical Volume Number": "E",
                "Logical Volume Access": "C",
                "Logical Volume Name": "A",
                "Logical Volume Size": "G",
                "Logical Volume Status": "D",
                "Major Device Number": "L",
                "Minor Device Number": "M",
                "Open Logical Volumes": "F",
                "Read Ahead Sectors": "K",
                "Volume Group Name": "B",
            }
        }


def test_pvcreate():
    """
    Tests for set a physical device to be used as an LVM physical volume
    """
    assert linux_lvm.pvcreate("") == "Error: at least one device is required"

    assert linux_lvm.pvcreate("A") == "A does not exist"

    # pvdisplay() would be called by pvcreate() twice: firstly to check
    # whether a device is already initialized for use by LVM and then to
    # ensure that the pvcreate executable did its job correctly.
    pvdisplay = MagicMock(side_effect=[False, True])
    with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
        with patch.object(os.path, "exists", return_value=True):
            ret = {
                "stdout": "saltines",
                "stderr": "cheese",
                "retcode": 0,
                "pid": "1337",
            }
            mock = MagicMock(return_value=ret)
            with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
                assert linux_lvm.pvcreate("A", metadatasize=1000) is True


def test_pvcreate_existing_pvs():
    """
    Test a scenario when all the submitted devices are already LVM PVs.
    """
    pvdisplay = MagicMock(return_value=True)
    with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
        with patch.object(os.path, "exists", return_value=True):
            ret = {
                "stdout": "saltines",
                "stderr": "cheese",
                "retcode": 0,
                "pid": "1337",
            }
            cmd_mock = MagicMock(return_value=ret)
            with patch.dict(linux_lvm.__salt__, {"cmd.run_all": cmd_mock}):
                assert linux_lvm.pvcreate("A", metadatasize=1000) is True
                assert cmd_mock.call_count == 0


def test_pvremove_not_pv():
    """
    Tests for remove a physical device not being used as an LVM physical volume
    """
    pvdisplay = MagicMock(return_value=False)
    with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
        assert linux_lvm.pvremove("A", override=False) == "A is not a physical volume"

    pvdisplay = MagicMock(return_value=False)
    with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
        assert linux_lvm.pvremove("A") is True


def test_pvremove():
    """
    Tests for remove a physical device being used as an LVM physical volume
    """
    pvdisplay = MagicMock(return_value=False)
    with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
        mock = MagicMock(return_value=True)
        with patch.dict(linux_lvm.__salt__, {"lvm.pvdisplay": mock}):
            ret = {
                "stdout": "saltines",
                "stderr": "cheese",
                "retcode": 0,
                "pid": "1337",
            }
            mock = MagicMock(return_value=ret)
            with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
                assert linux_lvm.pvremove("A") is True


def test_pvresize_not_pv():
    """
    Tests for resize a physical device not being used as an LVM physical volume
    """
    pvdisplay = MagicMock(return_value=False)
    with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
        assert linux_lvm.pvresize("A", override=False) == "A is not a physical volume"

    pvdisplay = MagicMock(return_value=False)
    with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
        assert linux_lvm.pvresize("A") is True


def test_pvresize():
    """
    Tests for resize a physical device being used as an LVM physical volume
    """
    pvdisplay = MagicMock(return_value=False)
    with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
        mock = MagicMock(return_value=True)
        with patch.dict(linux_lvm.__salt__, {"lvm.pvdisplay": mock}):
            ret = {
                "stdout": "saltines",
                "stderr": "cheese",
                "retcode": 0,
                "pid": "1337",
            }
            mock = MagicMock(return_value=ret)
            with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
                assert linux_lvm.pvresize("A") is True


def test_vgcreate():
    """
    Tests create an LVM volume group
    """
    assert linux_lvm.vgcreate("", "") == "Error: vgname and device(s) are both required"

    mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        with patch.object(linux_lvm, "vgdisplay", return_value={}):
            assert linux_lvm.vgcreate("fakevg", "B") == {
                "Output from vgcreate": ('Volume group "fakevg" successfully created')
            }


def test_vgextend():
    """
    Tests add physical volumes to an LVM volume group
    """
    assert linux_lvm.vgextend("", "") == "Error: vgname and device(s) are both required"

    mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        with patch.object(linux_lvm, "vgdisplay", return_value={}):
            assert linux_lvm.vgextend("fakevg", "B") == {
                "Output from vgextend": ('Volume group "fakevg" successfully extended')
            }


def test_lvcreate():
    """
    Test create a new logical volume, with option
    for which physical volume to be used
    """
    assert (
        linux_lvm.lvcreate(None, None, 1, 1)
        == "Error: Please specify only one of size or extents"
    )

    assert (
        linux_lvm.lvcreate(None, None, None, None)
        == "Error: Either size or extents must be specified"
    )

    assert (
        linux_lvm.lvcreate(None, None, thinvolume=True, thinpool=True)
        == "Error: Please set only one of thinvolume or thinpool to True"
    )

    assert (
        linux_lvm.lvcreate(None, None, thinvolume=True, extents=1)
        == "Error: Thin volume size cannot be specified as extents"
    )

    mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        with patch.object(linux_lvm, "lvdisplay", return_value={}):
            assert linux_lvm.lvcreate(None, None, None, 1) == {
                "Output from lvcreate": 'Logical volume "None" created.'
            }


def test_lvcreate_with_force():
    """
    Test create a new logical volume, with option
    for which physical volume to be used
    """
    mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        with patch.object(linux_lvm, "lvdisplay", return_value={}):
            assert linux_lvm.lvcreate(None, None, None, 1, force=True) == {
                "Output from lvcreate": 'Logical volume "None" created.'
            }


def test_lvcreate_extra_arguments_no_parameter():
    extra_args = {
        "nosync": None,
        "noudevsync": None,
        "ignoremonitoring": None,
        "thin": None,
    }
    mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        with patch.object(linux_lvm, "lvdisplay", return_value={}):
            assert linux_lvm.lvcreate(None, None, None, 1, **extra_args) == {
                "Output from lvcreate": 'Logical volume "None" created.'
            }
    expected_args = [f"--{arg}" for arg in extra_args]
    processed_extra_args = mock.call_args.args[0][-(len(extra_args) + 1) : -1]
    assert all([arg in expected_args for arg in processed_extra_args])


def test_lvcreate_invalid_extra_parameter():
    invalid_parameter = {"foo": "bar"}
    mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        with patch.object(linux_lvm, "lvdisplay", return_value={}):
            assert linux_lvm.lvcreate(None, None, None, 1, **invalid_parameter) == {
                "Output from lvcreate": 'Logical volume "None" created.'
            }
    processed_command = mock.call_args.args[0]
    assert "--foo" not in processed_command


def test_vgremove():
    """
    Tests to remove an LVM volume group
    """
    mock = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": '  Volume group "fakevg" successfully removed',
        }
    )
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert (
            linux_lvm.vgremove("fakevg") == 'Volume group "fakevg" successfully removed'
        )


def test_lvremove():
    """
    Test to remove a given existing logical volume
    from a named existing volume group
    """
    mock = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": '  Logical volume "lvtest" successfully removed',
        }
    )
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert (
            linux_lvm.lvremove("fakelv", "fakevg")
            == 'Logical volume "fakelv" successfully removed'
        )


def test_lvresize():
    """
    Tests to resize an LVM logical volume
    """
    assert linux_lvm.lvresize(1, None, 1) == {}

    assert linux_lvm.lvresize(None, None, None) == {}

    mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.lvresize(12, "/dev/fakevg/fakelv") == {
            "Output from lvresize": (
                'Logical volume "/dev/fakevg/fakelv" successfully resized.'
            )
        }


def test_lvextend():
    """
    Tests to extend an LVM logical volume
    """
    assert linux_lvm.lvextend(1, None, 1) == {}

    assert linux_lvm.lvextend(None, None, None) == {}

    mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
    with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
        assert linux_lvm.lvextend(12, "/dev/fakevg/fakelv") == {
            "Output from lvextend": (
                'Logical volume "/dev/fakevg/fakelv" successfully extended.'
            )
        }



"""
    tests.unit.test_virtualname
    ~~~~~~~~~~~~~~~~~~~~
"""

import importlib.util
import logging
import os

from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class FakeEntry:
    def __init__(self, name, path, is_file=True):
        self.name = name
        self.path = path
        self._is_file = is_file

    def is_file(self):
        return self._is_file


class VirtualNameTestCase(TestCase):
    """
    Test that the virtualname is in the module name, to speed up lookup of
    modules.
    """

    maxDiff = None

    @staticmethod
    def _import_module(testpath):
        spec = importlib.util.spec_from_file_location("tmpmodule", testpath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def _check_modules(self, path):
        """
        check modules in directory
        """
        ret = []
        for entry in os.listdir(path):
            name, path = os.path.splitext(os.path.basename(entry))[0], entry
            if name.startswith(".") or name.startswith("_"):
                continue
            if os.path.isfile(path) and not name.endswith(".py"):
                continue
            testpath = (
                path if os.path.isfile(path) else os.path.join(path, "__init__.py")
            )
            module = self._import_module(testpath)
            if hasattr(module, "__virtualname__"):
                if module.__virtualname__ not in name:
                    ret.append(
                        'Virtual name "{}" is not in the module filename "{}": {}'.format(
                            module.__virtualname__, name, path
                        )
                    )
        return ret

    def test_check_virtualname(self):
        """
        Test that the virtualname is in __name__ of the module
        """
        errors = []
        for entry in os.listdir(RUNTIME_VARS.SALT_CODE_DIR):
            name, path = os.path.splitext(os.path.basename(entry))[0], entry
            if name.startswith(".") or name.startswith("_") or not os.path.isdir(path):
                continue
            if name in ("cli", "defaults", "spm", "daemons", "ext", "templates"):
                continue
            if name == "cloud":
                entry = os.path.join(RUNTIME_VARS.SALT_CODE_DIR, "cloud", "clouds")
            errors.extend(self._check_modules(entry))
        for error in errors:
            log.critical(error)
        assert not errors



import pytest

from tests.support.case import ModuleCase


@pytest.mark.windows_whitelisted
@pytest.mark.timeout_unless_on_windows(120)
class DecoratorTest(ModuleCase):
    @pytest.mark.slow_test
    def test_module(self):
        self.assertTrue(self.run_function("runtests_decorators.working_function"))

    @pytest.mark.slow_test
    def test_depends(self):
        ret = self.run_function("runtests_decorators.depends")
        self.assertTrue(isinstance(ret, dict))
        self.assertTrue(ret["ret"])
        self.assertTrue(isinstance(ret["time"], float))

    @pytest.mark.slow_test
    def test_missing_depends(self):
        self.assertEqual(
            {
                "runtests_decorators.missing_depends_will_fallback": None,
                "runtests_decorators.missing_depends": (
                    "'runtests_decorators.missing_depends' is not available."
                ),
            },
            self.run_function("runtests_decorators.missing_depends"),
        )

    @pytest.mark.slow_test
    def test_bool_depends(self):
        # test True
        self.assertTrue(self.run_function("runtests_decorators.booldependsTrue"))

        # test False
        self.assertIn(
            "is not available",
            self.run_function("runtests_decorators.booldependsFalse"),
        )

    @pytest.mark.slow_test
    def test_depends_will_not_fallback(self):
        ret = self.run_function("runtests_decorators.depends_will_not_fallback")
        self.assertTrue(isinstance(ret, dict))
        self.assertTrue(ret["ret"])
        self.assertTrue(isinstance(ret["time"], float))

    @pytest.mark.slow_test
    def test_missing_depends_will_fallback(self):
        self.assertListEqual(
            [False, "fallback"],
            self.run_function("runtests_decorators.missing_depends_will_fallback"),
        )

    @pytest.mark.slow_test
    def test_command_success_retcode(self):
        ret = self.run_function("runtests_decorators.command_success_retcode")
        self.assertIs(ret, True)

    @pytest.mark.slow_test
    def test_command_failure_retcode(self):
        ret = self.run_function("runtests_decorators.command_failure_retcode")
        self.assertEqual(
            ret, "'runtests_decorators.command_failure_retcode' is not available."
        )

    @pytest.mark.slow_test
    def test_command_success_nonzero_retcode_true(self):
        ret = self.run_function(
            "runtests_decorators.command_success_nonzero_retcode_true"
        )
        self.assertIs(ret, True)

    @pytest.mark.slow_test
    def test_command_failure_nonzero_retcode_true(self):
        ret = self.run_function(
            "runtests_decorators.command_failure_nonzero_retcode_true"
        )
        self.assertEqual(
            ret,
            "'runtests_decorators.command_failure_nonzero_retcode_true' is not"
            " available.",
        )

    @pytest.mark.slow_test
    def test_command_success_nonzero_retcode_false(self):
        ret = self.run_function(
            "runtests_decorators.command_success_nonzero_retcode_false"
        )
        self.assertIs(ret, True)

    @pytest.mark.slow_test
    def test_command_failure_nonzero_retcode_false(self):
        ret = self.run_function(
            "runtests_decorators.command_failure_nonzero_retcode_false"
        )
        self.assertEqual(
            ret,
            "'runtests_decorators.command_failure_nonzero_retcode_false' is not"
            " available.",
        )

    @pytest.mark.slow_test
    def test_versioned_depend_insufficient(self):
        self.assertIn(
            "is not available",
            self.run_function("runtests_decorators.version_depends_false"),
        )

    @pytest.mark.slow_test
    def test_versioned_depend_sufficient(self):
        self.assertTrue(self.run_function("runtests_decorators.version_depends_true"))

    @pytest.mark.slow_test
    def test_versioned_depend_versionless(self):
        self.assertTrue(
            self.run_function("runtests_decorators.version_depends_versionless_true")
        )



"""
    Unit tests for the salt.modules.state module
"""

import salt.loader.context
import salt.modules.state
from tests.support.mock import patch


def test_get_initial_pillar():
    """
    _get_initial_pillar returns pillar data not named context
    """
    ctx = salt.loader.context.LoaderContext()
    pillar_data = {"foo": "bar"}
    named_ctx = ctx.named_context("__pillar__", pillar_data)
    opts = {"__cli": "salt-call", "pillarenv": "base"}
    with patch("salt.modules.state.__pillar__", named_ctx, create=True):
        with patch("salt.modules.state.__opts__", opts, create=True):
            pillar = salt.modules.state._get_initial_pillar(opts)
            assert pillar == pillar_data



import errno
import logging
import os
import shutil

import pytest

import salt.utils.files
from salt import fileclient
from tests.support.mock import patch

log = logging.getLogger(__name__)

SUBDIR = "subdir"


def _saltenvs():
    return ("base", "dev")


@pytest.fixture(params=_saltenvs())
def saltenv(request):
    return request.param


def _subdir_files():
    return ("foo.txt", "bar.txt", "baz.txt")


def _get_file_roots(fs_root):
    return {x: [os.path.join(fs_root, x)] for x in _saltenvs()}


@pytest.fixture
def fs_root(tmp_path):
    return os.path.join(tmp_path, "fileclient_fs_root")


@pytest.fixture
def cache_root(tmp_path):
    return os.path.join(tmp_path, "fileclient_cache_root")


@pytest.fixture
def mocked_opts(tmp_path, fs_root, cache_root):
    return {
        "file_roots": _get_file_roots(fs_root),
        "fileserver_backend": ["roots"],
        "cachedir": cache_root,
        "file_client": "local",
    }


@pytest.fixture
def configure_loader_modules(tmp_path, mocked_opts):
    return {fileclient: {"__opts__": mocked_opts}}


@pytest.fixture(autouse=True)
def _setup(fs_root, cache_root):
    """
    No need to add a dummy foo.txt to muddy up the github repo, just make
    our own fileserver root on-the-fly.
    """

    def _new_dir(path):
        """
        Add a new dir at ``path`` using os.makedirs. If the directory
        already exists, remove it recursively and then try to create it
        again.
        """
        try:
            os.makedirs(path)
        except OSError as exc:
            if exc.errno == errno.EEXIST:
                # Just in case a previous test was interrupted, remove the
                # directory and try adding it again.
                shutil.rmtree(path)
                os.makedirs(path)
            else:
                raise

    # Crete the FS_ROOT
    for saltenv in _saltenvs():
        saltenv_root = os.path.join(fs_root, saltenv)
        # Make sure we have a fresh root dir for this saltenv
        _new_dir(saltenv_root)

        path = os.path.join(saltenv_root, "foo.txt")
        with salt.utils.files.fopen(path, "w") as fp_:
            fp_.write(f"This is a test file in the '{saltenv}' saltenv.\n")

        subdir_abspath = os.path.join(saltenv_root, SUBDIR)
        os.makedirs(subdir_abspath)
        for subdir_file in _subdir_files():
            path = os.path.join(subdir_abspath, subdir_file)
            with salt.utils.files.fopen(path, "w") as fp_:
                fp_.write(
                    "This is file '{}' in subdir '{} from saltenv '{}'".format(
                        subdir_file, SUBDIR, saltenv
                    )
                )

    # Create the CACHE_ROOT
    _new_dir(cache_root)


@pytest.mark.parametrize("subdir", [SUBDIR, f"{SUBDIR}{os.sep}"])
def test_cache_dir(mocked_opts, minion_opts, subdir, saltenv):
    """
    Ensure entire directory is cached to correct location
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        assert client.cache_dir(
            f"salt://{subdir}",
            saltenv,
            exclude_pat="*foo.txt",
            cachedir=None,
        )
        for subdir_file in _subdir_files():
            cache_loc = os.path.join(
                fileclient.__opts__["cachedir"],
                "files",
                saltenv,
                subdir,
                subdir_file,
            )
            if not subdir_file.endswith("foo.txt"):
                with salt.utils.files.fopen(cache_loc) as fp_:
                    content = fp_.read()
                log.debug("cache_loc = %s", cache_loc)
                log.debug("content = %s", content)
                assert subdir_file in content
                assert SUBDIR in content
                assert saltenv in content
            else:
                assert not os.path.exists(cache_loc)


def test_cache_dir_include_empty(mocked_opts, minion_opts, fs_root):
    """
    Ensure entire directory is cached to correct location, including empty ones
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        for saltenv in _saltenvs():
            # Make an empty directory
            os.makedirs(os.path.join(fs_root, saltenv, SUBDIR, "emptydir"))
            os.makedirs(os.path.join(fs_root, saltenv, SUBDIR, "emptydir2"))
            os.makedirs(
                os.path.join(
                    fileclient.__opts__["cachedir"],
                    "files",
                    saltenv,
                    SUBDIR,
                    "emptydir2",
                )
            )
            os.makedirs(os.path.join(fs_root, saltenv, "emptydir"))
            assert client.cache_dir(
                f"salt://{SUBDIR}", saltenv, cachedir=None, include_empty=True
            )
            for subdir_file in _subdir_files():
                cache_loc = os.path.join(
                    fileclient.__opts__["cachedir"],
                    "files",
                    saltenv,
                    SUBDIR,
                    subdir_file,
                )
                with salt.utils.files.fopen(cache_loc) as fp_:
                    content = fp_.read()
                log.debug("cache_loc = %s", cache_loc)
                log.debug("content = %s", content)
                assert subdir_file in content
                assert SUBDIR in content
                assert saltenv in content
            assert os.path.exists(
                os.path.join(
                    fileclient.__opts__["cachedir"],
                    "files",
                    saltenv,
                    SUBDIR,
                    "emptydir",
                )
            )
            assert not os.path.exists(
                os.path.join(
                    fileclient.__opts__["cachedir"], "files", saltenv, "emptydir"
                )
            )


def test_cache_dir_with_alternate_cachedir_and_absolute_path(
    mocked_opts, minion_opts, tmp_path
):
    """
    Ensure entire directory is cached to correct location when an alternate
    cachedir is specified and that cachedir is an absolute path
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)
    alt_cachedir = os.path.join(tmp_path, "abs_cachedir")

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        for saltenv in _saltenvs():
            assert client.cache_dir(f"salt://{SUBDIR}", saltenv, cachedir=alt_cachedir)
            for subdir_file in _subdir_files():
                cache_loc = os.path.join(
                    alt_cachedir, "files", saltenv, SUBDIR, subdir_file
                )
                # Double check that the content of the cached file
                # identifies it as being from the correct saltenv. The
                # setUp function creates the file with the name of the
                # saltenv mentioned in the file, so a simple 'in' check is
                # sufficient here. If opening the file raises an exception,
                # this is a problem, so we are not catching the exception
                # and letting it be raised so that the test fails.
                with salt.utils.files.fopen(cache_loc) as fp_:
                    content = fp_.read()
                log.debug("cache_loc = %s", cache_loc)
                log.debug("content = %s", content)
                assert subdir_file in content
                assert SUBDIR in content
                assert saltenv in content


def test_cache_dir_with_alternate_cachedir_and_relative_path(mocked_opts, minion_opts):
    """
    Ensure entire directory is cached to correct location when an alternate
    cachedir is specified and that cachedir is a relative path
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)
    alt_cachedir = "foo"

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        for saltenv in _saltenvs():
            assert client.cache_dir(f"salt://{SUBDIR}", saltenv, cachedir=alt_cachedir)
            for subdir_file in _subdir_files():
                cache_loc = os.path.join(
                    fileclient.__opts__["cachedir"],
                    alt_cachedir,
                    "files",
                    saltenv,
                    SUBDIR,
                    subdir_file,
                )
                # Double check that the content of the cached file
                # identifies it as being from the correct saltenv. The
                # setUp function creates the file with the name of the
                # saltenv mentioned in the file, so a simple 'in' check is
                # sufficient here. If opening the file raises an exception,
                # this is a problem, so we are not catching the exception
                # and letting it be raised so that the test fails.
                with salt.utils.files.fopen(cache_loc) as fp_:
                    content = fp_.read()
                log.debug("cache_loc = %s", cache_loc)
                log.debug("content = %s", content)
                assert subdir_file in content
                assert SUBDIR in content
                assert saltenv in content


def test_cache_file(mocked_opts, minion_opts):
    """
    Ensure file is cached to correct location
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        for saltenv in _saltenvs():
            assert client.cache_file("salt://foo.txt", saltenv, cachedir=None)
            cache_loc = os.path.join(
                fileclient.__opts__["cachedir"], "files", saltenv, "foo.txt"
            )
            # Double check that the content of the cached file identifies
            # it as being from the correct saltenv. The setUp function
            # creates the file with the name of the saltenv mentioned in
            # the file, so a simple 'in' check is sufficient here. If
            # opening the file raises an exception, this is a problem, so
            # we are not catching the exception and letting it be raised so
            # that the test fails.
            with salt.utils.files.fopen(cache_loc) as fp_:
                content = fp_.read()
            log.debug("cache_loc = %s", cache_loc)
            log.debug("content = %s", content)
            assert saltenv in content


def test_cache_files(mocked_opts, minion_opts):
    """
    Test caching multiple files
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        files = ["salt://foo.txt", f"salt://{SUBDIR}/bar.txt"]
        for saltenv in _saltenvs():
            assert client.cache_files(files, saltenv, cachedir=None)
            for fn in files:
                fn = fn[7:]
                cache_loc = os.path.join(
                    fileclient.__opts__["cachedir"], "files", saltenv, fn
                )
                with salt.utils.files.fopen(cache_loc) as fp_:
                    content = fp_.read()
                log.debug("cache_loc = %s", cache_loc)
                log.debug("content = %s", content)
                assert saltenv in content


def test_cache_files_string(mocked_opts, minion_opts):
    """
    Test caching multiple files
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        files = ["salt://foo.txt", f"salt://{SUBDIR}/bar.txt"]
        for saltenv in _saltenvs():
            assert client.cache_files(",".join(files), saltenv, cachedir=None)
            for fn in files:
                fn = fn[7:]
                cache_loc = os.path.join(
                    fileclient.__opts__["cachedir"], "files", saltenv, fn
                )
                with salt.utils.files.fopen(cache_loc) as fp_:
                    content = fp_.read()
                log.debug("cache_loc = %s", cache_loc)
                log.debug("content = %s", content)
                assert saltenv in content


def test_cache_master(mocked_opts, minion_opts, fs_root):
    """
    Test caching multiple files
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        for saltenv in _saltenvs():
            assert client.cache_master(saltenv=saltenv)
            for dirpath, _, filenames in os.walk(os.path.join(fs_root, saltenv)):
                for fn in filenames:
                    fileserver_path = os.path.relpath(
                        os.path.join(dirpath, fn), fs_root
                    )
                    cachedir_path = os.path.join(
                        fileclient.__opts__["cachedir"], "files", fileserver_path
                    )
                    assert os.path.exists(cachedir_path)
                    with salt.utils.files.fopen(cachedir_path) as fp_:
                        content = fp_.read()
                    log.debug("cache_loc = %s", cachedir_path)
                    log.debug("content = %s", content)
                    assert saltenv in content


def test_cache_file_with_alternate_cachedir_and_absolute_path(
    mocked_opts, minion_opts, tmp_path
):
    """
    Ensure file is cached to correct location when an alternate cachedir is
    specified and that cachedir is an absolute path
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)
    alt_cachedir = os.path.join(tmp_path, "abs_cachedir")

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        for saltenv in _saltenvs():
            assert client.cache_file("salt://foo.txt", saltenv, cachedir=alt_cachedir)
            cache_loc = os.path.join(alt_cachedir, "files", saltenv, "foo.txt")
            # Double check that the content of the cached file identifies
            # it as being from the correct saltenv. The setUp function
            # creates the file with the name of the saltenv mentioned in
            # the file, so a simple 'in' check is sufficient here. If
            # opening the file raises an exception, this is a problem, so
            # we are not catching the exception and letting it be raised so
            # that the test fails.
            with salt.utils.files.fopen(cache_loc) as fp_:
                content = fp_.read()
            log.debug("cache_loc = %s", cache_loc)
            log.debug("content = %s", content)
            assert saltenv in content


def test_cache_file_with_alternate_cachedir_and_relative_path(mocked_opts, minion_opts):
    """
    Ensure file is cached to correct location when an alternate cachedir is
    specified and that cachedir is a relative path
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)
    alt_cachedir = "foo"

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        for saltenv in _saltenvs():
            assert client.cache_file("salt://foo.txt", saltenv, cachedir=alt_cachedir)
            cache_loc = os.path.join(
                fileclient.__opts__["cachedir"],
                alt_cachedir,
                "files",
                saltenv,
                "foo.txt",
            )
            # Double check that the content of the cached file identifies
            # it as being from the correct saltenv. The setUp function
            # creates the file with the name of the saltenv mentioned in
            # the file, so a simple 'in' check is sufficient here. If
            # opening the file raises an exception, this is a problem, so
            # we are not catching the exception and letting it be raised so
            # that the test fails.
            with salt.utils.files.fopen(cache_loc) as fp_:
                content = fp_.read()
            log.debug("cache_loc = %s", cache_loc)
            log.debug("content = %s", content)
            assert saltenv in content


def test_cache_dest(mocked_opts, minion_opts):
    """
    Tests functionality for cache_dest
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)

    relpath = "foo.com/bar.txt"
    cachedir = minion_opts["cachedir"]

    def _external(saltenv="base"):
        return salt.utils.path.join(
            patched_opts["cachedir"], "extrn_files", saltenv, relpath
        )

    def _salt(saltenv="base"):
        return salt.utils.path.join(patched_opts["cachedir"], "files", saltenv, relpath)

    def _check(ret, expected):
        assert ret == expected, f"{ret} != {expected}"

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)

        _check(client.cache_dest(f"https://{relpath}"), _external())

        _check(client.cache_dest(f"https://{relpath}", "dev"), _external("dev"))

        _check(client.cache_dest(f"salt://{relpath}"), _salt())

        _check(client.cache_dest(f"salt://{relpath}", "dev"), _salt("dev"))

        _check(client.cache_dest(f"salt://{relpath}?saltenv=dev"), _salt("dev"))

        _check(client.cache_dest("/foo/bar"), "/foo/bar")


def test_is_cached_salt_proto_saltenv(mocked_opts, minion_opts):
    """
    Test that is_cached works with a salt protocol filepath and a saltenv query param
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        file_url = "salt://foo.txt?saltenv=dev"
        client.cache_file(file_url)
        assert client.is_cached(file_url)


def test_is_cached_local_files(mocked_opts, minion_opts):
    """
    Test that is_cached works with local files
    """
    patched_opts = minion_opts.copy()
    patched_opts.update(mocked_opts)

    with patch.dict(fileclient.__opts__, patched_opts):
        client = fileclient.get_file_client(fileclient.__opts__, pillar=False)
        local_file = os.path.join(
            patched_opts["cachedir"], "localfiles", "testfile.txt"
        )
        os.makedirs(os.path.join(patched_opts["cachedir"], "localfiles"), exist_ok=True)
        with salt.utils.files.fopen(local_file, "w") as fp:
            fp.write("test")
        assert client.is_cached("testfile.txt")



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest

import salt.states.locale as locale
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {locale: {}}


def test_system():
    """
    Test to set the locale for the system
    """
    ret = [
        {
            "changes": {},
            "comment": "System locale salt already set",
            "name": "salt",
            "result": True,
        },
        {
            "changes": {},
            "comment": "System locale saltstack needs to be set",
            "name": "saltstack",
            "result": None,
        },
        {
            "changes": {"locale": "saltstack"},
            "comment": "Set system locale saltstack",
            "name": "saltstack",
            "result": True,
        },
        {
            "changes": {},
            "comment": "Failed to set system locale to saltstack",
            "name": "saltstack",
            "result": False,
        },
    ]

    mock = MagicMock(return_value="salt")
    with patch.dict(locale.__salt__, {"locale.get_locale": mock}):
        assert locale.system("salt") == ret[0]

        with patch.dict(locale.__opts__, {"test": True}):
            assert locale.system("saltstack") == ret[1]

        with patch.dict(locale.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[True, False])
            with patch.dict(locale.__salt__, {"locale.set_locale": mock}):
                assert locale.system("saltstack") == ret[2]

                assert locale.system("saltstack") == ret[3]


def test_present():
    """
    Test to generate a locale if it is not present
    """
    ret = [
        {
            "changes": {},
            "comment": "Locale salt is already present",
            "name": "salt",
            "result": True,
        },
        {
            "changes": {},
            "comment": "Locale salt needs to be generated",
            "name": "salt",
            "result": None,
        },
        {
            "changes": {"locale": "salt"},
            "comment": "Generated locale salt",
            "name": "salt",
            "result": True,
        },
        {
            "changes": {},
            "comment": "Failed to generate locale salt",
            "name": "salt",
            "result": False,
        },
    ]

    mock = MagicMock(side_effect=[True, False, False, False])
    with patch.dict(locale.__salt__, {"locale.avail": mock}):
        assert locale.present("salt") == ret[0]

        with patch.dict(locale.__opts__, {"test": True}):
            assert locale.present("salt") == ret[1]

        with patch.dict(locale.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[True, False])
            with patch.dict(locale.__salt__, {"locale.gen_locale": mock}):
                assert locale.present("salt") == ret[2]

                assert locale.present("salt") == ret[3]



import logging

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("ipset", check_all=False),
]


@pytest.fixture()
def setup_set(modules):
    set_name = "test_name"
    kwargs = {"range": "192.168.0.0/16", "comment": "Hello18"}
    modules.ipset.new_set(name=set_name, set_type="bitmap:ip", family="ipv4", **kwargs)
    yield set_name
    modules.ipset.delete_set(set_name)


def test_ipset_present(states, setup_set):
    """
    test ipset.present
    """
    # add set first
    entry = "192.168.0.3"
    comment = "Hello"
    ret = states.ipset.present(
        name="setname_entries", set_name=setup_set, entry=entry, comment=comment
    )
    assert ret.result
    assert (
        ret.comment
        == f"entry {entry} comment {comment} added to set {setup_set} for family ipv4\n"
    )



import pytest

import salt.utils.win_network as win_network
from tests.support.mock import MagicMock, patch

pytestmark = [pytest.mark.skip_unless_on_windows]


class PhysicalAddress:
    def __init__(self, address):
        self.address = address

    def ToString(self):
        return str(self.address)


class Interface:
    """
    Mocked interface object
    """

    def __init__(
        self,
        i_address="02D5F1DD31E0",
        i_description="Dell GigabitEthernet",
        i_id="{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
        i_name="Ethernet",
        i_receive_only=False,
        i_status=1,
        i_type=6,
    ):
        self.PhysicalAddress = PhysicalAddress(i_address)
        self.Description = i_description
        self.Id = i_id
        self.Name = i_name
        self.NetworkInterfaceType = i_type
        self.IsReceiveOnly = i_receive_only
        self.OperationalStatus = i_status

    def GetPhysicalAddress(self):
        return self.PhysicalAddress


@pytest.fixture
def mock_ip_base():
    return MagicMock(
        return_value={
            "dns_enabled": False,
            "dns_suffix": "",
            "dynamic_dns_enabled": False,
            "ipv4_apipa_active": False,
            "ipv4_apipa_enabled": True,
            "ipv4_dhcp_enabled": False,
            "ipv4_forwarding_enabled": False,
            "ipv4_index": 1,
            "ipv4_mtu": 1500,
            "ipv4_wins_enabled": False,
            "ipv6_index": 1,
            "ipv6_mtu": 1500,
        }
    )


@pytest.fixture
def mock_unicast():
    return MagicMock(
        return_value={
            "ip_addresses": [
                {
                    "address": "172.18.87.49",
                    "broadcast": "172.18.87.63",
                    "loopback": "127.0.0.1",
                    "netmask": "255.255.255.240",
                    "prefix_length": 28,
                    "prefix_origin": "Manual",
                    "suffix_origin": "Manual",
                }
            ],
            "ipv6_addresses": [
                {
                    "address": "fe80::e8a4:1224:5548:2b81",
                    "interface_index": 32,
                    "prefix_length": 64,
                    "prefix_origin": "WellKnown",
                    "suffix_origin": "Router",
                }
            ],
        }
    )


@pytest.fixture
def mock_gateway():
    return MagicMock(
        return_value={
            "ip_gateways": ["192.168.0.1"],
            "ipv6_gateways": ["fe80::208:a2ff:fe0b:de70"],
        }
    )


@pytest.fixture
def mock_dns():
    return MagicMock(
        return_value={
            "ip_dns": ["10.4.0.1", "10.1.0.1", "8.8.8.8"],
            "ipv6_dns": ["2600:740a:1:304::1"],
        }
    )


@pytest.fixture
def mock_multicast():
    return MagicMock(
        return_value={
            "ip_multicast": [
                "224.0.0.1",
                "224.0.0.251",
                "224.0.0.252",
                "230.230.230.230",
                "239.0.0.250",
                "239.255.255.250",
            ],
            "ipv6_multicast": [
                "ff01::1",
                "ff02::1",
                "ff02::c",
                "ff02::fb",
                "ff02::1:3",
                "ff02::1:ff0f:4c48",
                "ff02::1:ffa6:f6e6",
            ],
        }
    )


@pytest.fixture
def mock_anycast():
    return MagicMock(return_value={"ip_anycast": [], "ipv6_anycast": []})


@pytest.fixture
def mock_wins():
    return MagicMock(return_value={"ip_wins": []})


def test_get_interface_info_dot_net(
    mock_ip_base,
    mock_unicast,
    mock_gateway,
    mock_dns,
    mock_multicast,
    mock_anycast,
    mock_wins,
):
    expected = {
        "Ethernet": {
            "alias": "Ethernet",
            "description": "Dell GigabitEthernet",
            "dns_enabled": False,
            "dns_suffix": "",
            "dynamic_dns_enabled": False,
            "id": "{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
            "ip_addresses": [
                {
                    "address": "172.18.87.49",
                    "broadcast": "172.18.87.63",
                    "loopback": "127.0.0.1",
                    "netmask": "255.255.255.240",
                    "prefix_length": 28,
                    "prefix_origin": "Manual",
                    "suffix_origin": "Manual",
                }
            ],
            "ip_anycast": [],
            "ip_dns": ["10.4.0.1", "10.1.0.1", "8.8.8.8"],
            "ip_gateways": ["192.168.0.1"],
            "ip_multicast": [
                "224.0.0.1",
                "224.0.0.251",
                "224.0.0.252",
                "230.230.230.230",
                "239.0.0.250",
                "239.255.255.250",
            ],
            "ip_wins": [],
            "ipv4_apipa_active": False,
            "ipv4_apipa_enabled": True,
            "ipv4_dhcp_enabled": False,
            "ipv4_forwarding_enabled": False,
            "ipv4_index": 1,
            "ipv4_mtu": 1500,
            "ipv4_wins_enabled": False,
            "ipv6_addresses": [
                {
                    "address": "fe80::e8a4:1224:5548:2b81",
                    "interface_index": 32,
                    "prefix_length": 64,
                    "prefix_origin": "WellKnown",
                    "suffix_origin": "Router",
                }
            ],
            "ipv6_anycast": [],
            "ipv6_dns": ["2600:740a:1:304::1"],
            "ipv6_gateways": ["fe80::208:a2ff:fe0b:de70"],
            "ipv6_index": 1,
            "ipv6_mtu": 1500,
            "ipv6_multicast": [
                "ff01::1",
                "ff02::1",
                "ff02::c",
                "ff02::fb",
                "ff02::1:3",
                "ff02::1:ff0f:4c48",
                "ff02::1:ffa6:f6e6",
            ],
            "physical_address": "02:D5:F1:DD:31:E0",
            "receive_only": False,
            "status": "Up",
            "type": "Ethernet",
        }
    }

    mock_int = MagicMock(return_value=[Interface()])
    with patch.object(win_network, "_get_network_interfaces", mock_int), patch.object(
        win_network, "_get_ip_base_properties", mock_ip_base
    ), patch.object(win_network, "_get_ip_unicast_info", mock_unicast), patch.object(
        win_network, "_get_ip_gateway_info", mock_gateway
    ), patch.object(
        win_network, "_get_ip_dns_info", mock_dns
    ), patch.object(
        win_network, "_get_ip_multicast_info", mock_multicast
    ), patch.object(
        win_network, "_get_ip_anycast_info", mock_anycast
    ), patch.object(
        win_network, "_get_ip_wins_info", mock_wins
    ):

        results = win_network.get_interface_info_dot_net()

    assert expected == results


def test_get_network_info(
    mock_ip_base,
    mock_unicast,
    mock_gateway,
    mock_dns,
    mock_multicast,
    mock_anycast,
    mock_wins,
):
    expected = {
        "Ethernet": {
            "description": "Dell GigabitEthernet",
            "hwaddr": "02:D5:F1:DD:31:E0",
            "inet": [
                {
                    "address": "172.18.87.49",
                    "apipa_active": False,
                    "apipa_enabled": True,
                    "broadcast": "172.18.87.63",
                    "dhcp_enabled": False,
                    "forwarding_enabled": False,
                    "gateway": "192.168.0.1",
                    "index": 1,
                    "mtu": 1500,
                    "netmask": "255.255.255.240",
                    "wins_enabled": False,
                }
            ],
            "inet6": [
                {
                    "address": "fe80::e8a4:1224:5548:2b81",
                    "gateway": "fe80::208:a2ff:fe0b:de70",
                    "index": 1,
                    "mtu": 1500,
                    "prefixlen": 64,
                }
            ],
            "up": True,
        }
    }
    mock_int = MagicMock(return_value=[Interface()])
    with patch.object(win_network, "_get_network_interfaces", mock_int), patch.object(
        win_network, "_get_ip_base_properties", mock_ip_base
    ), patch.object(win_network, "_get_ip_unicast_info", mock_unicast), patch.object(
        win_network, "_get_ip_gateway_info", mock_gateway
    ), patch.object(
        win_network, "_get_ip_dns_info", mock_dns
    ), patch.object(
        win_network, "_get_ip_multicast_info", mock_multicast
    ), patch.object(
        win_network, "_get_ip_anycast_info", mock_anycast
    ), patch.object(
        win_network, "_get_ip_wins_info", mock_wins
    ):

        results = win_network.get_interface_info()

    assert expected == results


def test__get_base_properties_tap_adapter():
    """
    Adapter Type 53 is apparently an undocumented type corresponding to
    OpenVPN TAP Adapters and possibly other TAP Adapters. This test makes
    sure the win_network util will catch that.
    https://github.com/saltstack/salt/issues/56196
    https://github.com/saltstack/salt/issues/56275
    """
    i_face = Interface(
        i_address="03DE4D0713FA",
        i_description="Windows TAP Adapter",
        i_id="{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
        i_name="Windows TAP Adapter",
        i_receive_only=False,
        i_status=1,
        i_type=53,
    )
    expected = {
        "alias": "Windows TAP Adapter",
        "description": "Windows TAP Adapter",
        "id": "{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
        "receive_only": False,
        "physical_address": "03:DE:4D:07:13:FA",
        "status": "Up",
        "type": "TAPAdapter",
    }
    results = win_network._get_base_properties(i_face=i_face)
    assert expected == results


def test__get_base_properties_undefined_adapter():
    """
    The Adapter Type 53 may be an arbitrary number assigned by OpenVPN.
    This will test the ability to avoid stack tracing on an undefined
    adapter type. If one is encountered, just use the description.
    """
    i_face = Interface(
        i_address="03DE4D0713FA",
        i_description="Undefined Adapter",
        i_id="{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
        i_name="Undefined",
        i_receive_only=False,
        i_status=1,
        i_type=50,
    )
    expected = {
        "alias": "Undefined",
        "description": "Undefined Adapter",
        "id": "{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
        "receive_only": False,
        "physical_address": "03:DE:4D:07:13:FA",
        "status": "Up",
        "type": "Undefined Adapter",
    }
    results = win_network._get_base_properties(i_face=i_face)
    assert expected == results


def test__get_network_interfaces_no_error():
    ret = win_network._get_network_interfaces()
    assert len(ret) > 0



import os
import sys

import pytest

from salt._logging import get_log_record_factory, set_log_record_factory
from salt.cli.api import SaltAPI


@pytest.mark.slow_test
def test_start_shutdown(monkeypatch, tmp_path):
    pid_file = str(tmp_path / "pid_file")
    log_file = str(tmp_path / "log_file")
    api = SaltAPI()
    orig_factory = get_log_record_factory()
    with pytest.raises(SystemExit):
        # testing environment will fail if we use default pidfile
        # overwrite sys.argv so salt-api does not use testing args
        monkeypatch.setattr(
            "sys.argv", [sys.argv[0], "--pid-file", pid_file, "--log-file", log_file]
        )
        try:
            api.start()
            assert os.path.isfile(pid_file)
            assert os.path.isfile(log_file)
            api.shutdown()
        finally:
            set_log_record_factory(orig_factory)



"""
Management of Microsoft SQLServer Logins
========================================

The mssql_login module is used to create
and manage SQL Server Logins

.. code-block:: yaml

    frank:
      mssql_login.present
        - domain: mydomain
"""

import collections


def __virtual__():
    """
    Only load if the mssql module is present
    """
    if "mssql.version" in __salt__:
        return True
    return (False, "mssql module could not be loaded")


def _normalize_options(options):
    if type(options) in [dict, collections.OrderedDict]:
        return [f"{k}={v}" for k, v in options.items()]
    if type(options) is list and (not options or type(options[0]) is str):
        return options
    # Invalid options
    if type(options) is not list or type(options[0]) not in [
        dict,
        collections.OrderedDict,
    ]:
        return []
    return [o for d in options for o in _normalize_options(d)]


def present(
    name, password=None, domain=None, server_roles=None, options=None, **kwargs
):
    """
    Checks existence of the named login.
    If not present, creates the login with the specified roles and options.

    name
        The name of the login to manage
    password
        Creates a SQL Server authentication login
        Since hashed passwords are varbinary values, if the
        new_login_password is 'long', it will be considered
        to be HASHED.
    domain
        Creates a Windows authentication login.
        Needs to be NetBIOS domain or hostname
    server_roles
        Add this login to all the server roles in the list
    options
        Can be a list of strings, a dictionary, or a list of dictionaries
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if bool(password) == bool(domain):
        ret["result"] = False
        ret["comment"] = "One and only one of password and domain should be specifies"
        return ret
    if __salt__["mssql.login_exists"](name, domain=domain, **kwargs):
        ret["comment"] = (
            "Login {} is already present (Not going to try to set its password)".format(
                name
            )
        )
        return ret
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Login {name} is set to be added"
        return ret

    login_created = __salt__["mssql.login_create"](
        name,
        new_login_password=password,
        new_login_domain=domain,
        new_login_roles=server_roles,
        new_login_options=_normalize_options(options),
        **kwargs,
    )
    # Non-empty strings are also evaluated to True, so we cannot use if not login_created:
    if login_created is not True:
        ret["result"] = False
        ret["comment"] = f"Login {name} failed to be added: {login_created}"
        return ret
    ret["comment"] = f"Login {name} has been added. "
    ret["changes"][name] = "Present"
    return ret


def absent(name, **kwargs):
    """
    Ensure that the named login is absent

    name
        The name of the login to remove
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if not __salt__["mssql.login_exists"](name):
        ret["comment"] = f"Login {name} is not present"
        return ret
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Login {name} is set to be removed"
        return ret
    if __salt__["mssql.login_remove"](name, **kwargs):
        ret["comment"] = f"Login {name} has been removed"
        ret["changes"][name] = "Absent"
        return ret
    # else:
    ret["result"] = False
    ret["comment"] = f"Login {name} failed to be removed"
    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.ret
"""

import pytest

import salt.loader
import salt.modules.ret as ret
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {ret: {}}


# 'get_jid' function tests: 1


def test_get_jid():
    """
    Test if it return the information for a specified job id
    """
    mock_ret = MagicMock(return_value="DB")
    with patch.object(
        salt.loader,
        "returners",
        MagicMock(return_value={"redis.get_jid": mock_ret}),
    ):
        assert ret.get_jid("redis", "net") == "DB"


# 'get_fun' function tests: 1


def test_get_fun():
    """
    Test if it return info about last time fun was called on each minion
    """
    mock_ret = MagicMock(return_value="DB")
    with patch.object(
        salt.loader,
        "returners",
        MagicMock(return_value={"mysql.get_fun": mock_ret}),
    ):
        assert ret.get_fun("mysql", "net") == "DB"


# 'get_jids' function tests: 1


def test_get_jids():
    """
    Test if it return a list of all job ids
    """
    mock_ret = MagicMock(return_value="DB")
    with patch.object(
        salt.loader,
        "returners",
        MagicMock(return_value={"mysql.get_jids": mock_ret}),
    ):
        assert ret.get_jids("mysql") == "DB"


# 'get_minions' function tests: 1


def test_get_minions():
    """
    Test if it return a list of all minions
    """
    mock_ret = MagicMock(return_value="DB")
    with patch.object(
        salt.loader,
        "returners",
        MagicMock(return_value={"mysql.get_minions": mock_ret}),
    ):
        assert ret.get_minions("mysql") == "DB"



import logging

import pytest

import salt.cache
from tests.pytests.functional.cache.helpers import run_common_cache_tests

log = logging.getLogger(__name__)


@pytest.fixture
def cache(minion_opts):
    opts = minion_opts.copy()
    opts["memcache_expire_seconds"] = 42
    cache = salt.cache.factory(opts)
    return cache


def test_caching(subtests, cache):
    run_common_cache_tests(subtests, cache)



import logging

import salt.channel.client
import salt.config
import salt.crypt
import salt.utils.args
import salt.utils.jid

log = logging.getLogger(__name__)


def test_minoin_event_blacklist(salt_master, salt_minion, salt_cli, caplog):
    ret = salt_cli.run("test.ping", minion_tgt=salt_minion.id)
    assert ret.returncode == 0

    opts = salt.config.minion_config(salt_minion.config_file)
    opts["master_uri"] = "tcp://{}:{}".format(opts["master"], opts["master_port"])

    jid = salt.utils.jid.gen_jid(opts)
    auth = salt.crypt.SAuth(opts)
    tok = auth.gen_token(b"salt")

    load = {
        "cmd": "_minion_event",
        "tok": tok,
        "id": opts["id"],
        "events": [
            {
                "data": {
                    "fun": "test.ping",
                    "arg": [],
                    "jid": jid,
                    "ret": "",
                    "tgt": salt_minion.id,
                    "tgt_type": "glob",
                    "user": "root",
                    "__peer_id": "salt",
                },
                "tag": f"salt/job/{jid}/publish",
            }
        ],
    }
    with caplog.at_level(logging.WARNING):
        with salt.channel.client.ReqChannel.factory(opts) as channel:
            channel.send(load, tries=1, timeout=10000)
            log.info("payload sent, jid was %s", jid)
        assert "Filtering blacklisted" in caplog.text



"""Test minion configuration option startup_states.

There are four valid values for this option, which are validated by checking the jobs
executed after minion start.
"""

import pytest


@pytest.fixture
def salt_minion_startup_states_empty_string(salt_master, salt_minion_id):
    config_overrides = {
        "startup_states": "",
    }
    factory = salt_master.salt_minion_daemon(
        f"{salt_minion_id}-empty-string",
        overrides=config_overrides,
    )
    with factory.started():
        yield factory


@pytest.fixture
def salt_minion_startup_states_highstate(salt_master, salt_minion_id):
    config_overrides = {
        "startup_states": "highstate",
    }
    factory = salt_master.salt_minion_daemon(
        f"{salt_minion_id}-highstate",
        overrides=config_overrides,
    )
    with factory.started():
        yield factory


@pytest.fixture
def salt_minion_startup_states_sls(salt_master, salt_minion_id):
    config_overrides = {"startup_states": "sls", "sls_list": ["example-sls"]}
    factory = salt_master.salt_minion_daemon(
        f"{salt_minion_id}-sls",
        overrides=config_overrides,
    )
    with factory.started():
        yield factory


@pytest.fixture
def salt_minion_startup_states_top(salt_master, salt_minion_id):
    config_overrides = {"startup_states": "top", "top_file": "example-top.sls"}
    factory = salt_master.salt_minion_daemon(
        f"{salt_minion_id}-top",
        overrides=config_overrides,
    )
    with factory.started():
        yield factory


def test_startup_states_empty_string(
    salt_run_cli, salt_minion_startup_states_empty_string
):
    # Get jobs for this minion
    ret = salt_run_cli.run(
        "jobs.list_jobs", f"search_target={salt_minion_startup_states_empty_string.id}"
    )
    # Check no job was run
    assert len(ret.data.keys()) == 0


def test_startup_states_highstate(salt_run_cli, salt_minion_startup_states_highstate):
    with salt_minion_startup_states_highstate:
        # Get jobs for this minion
        ret = salt_run_cli.run(
            "jobs.list_jobs", f"search_target={salt_minion_startup_states_highstate.id}"
        )
        # Check there is exactly one job
        assert len(ret.data.keys()) == 1
        # Check that job executes state.highstate
        job_ret = next(iter(ret.data.values()))
        assert "Function" in job_ret
        assert job_ret["Function"] == "state.highstate"
        assert "Arguments" in job_ret
        assert job_ret["Arguments"] == []


def test_startup_states_sls(salt_run_cli, salt_minion_startup_states_sls):
    with salt_minion_startup_states_sls:
        # Get jobs for this minion
        ret = salt_run_cli.run(
            "jobs.list_jobs", f"search_target={salt_minion_startup_states_sls.id}"
        )
        # Check there is exactly one job
        assert len(ret.data.keys()) == 1
        # Check that job executes state.sls
        job_ret = next(iter(ret.data.values()))
        assert "Function" in job_ret
        assert job_ret["Function"] == "state.sls"
        assert "Arguments" in job_ret
        assert job_ret["Arguments"] == [["example-sls"]]


def test_startup_states_top(salt_run_cli, salt_minion_startup_states_top):
    with salt_minion_startup_states_top:
        # Get jobs for this minion
        ret = salt_run_cli.run(
            "jobs.list_jobs", f"search_target={salt_minion_startup_states_top.id}"
        )
        # Check there is exactly one job
        assert len(ret.data.keys()) == 1
        # Check that job executes state.top
        job_ret = next(iter(ret.data.values()))
        assert "Function" in job_ret
        assert job_ret["Function"] == "state.top"
        assert "Arguments" in job_ret
        assert job_ret["Arguments"] == ["example-top.sls"]
